#!/usr/bin/env python
# cardinal_pythonlib/psychiatry/drugs.py

"""
===============================================================================

    Original code copyright (C) 2009-2021 Rudolf Cardinal (rudolf@pobox.com).

    This file is part of cardinal_pythonlib.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        https://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

===============================================================================

**Drug information, with an emphasis on psychotropic drugs, including
translating specific to generic names.**

**Examples**

Test within Python:

.. code-block:: python

    from cardinal_pythonlib.psychiatry.drugs import *

    drug_name_to_generic("UNKNOWN")
    drug_name_to_generic("UNKNOWN", unknown_to_default=True)
    drug_names_to_generic([
        "citalopram", "Citalopram", "Cipramil", "Celexa",
        "olanzepine",  # typo
        "dextroamphetamine",
        "amitryptyline",
    ])

**Antidepressants**

As of 2018-07-01, this is a functional superset of the SLAM
antidepressant-finding SQL (see ``dep_med_v1``), though mainly a superset in
non-antidepressant respects; the only antidepressants it adds are:

- buproprion, maprotiline

The SLAM antidepressant finder finds:

- tricyclic (category)
- amitriptyline, clomipramine, dosulepin, doxepin, imipramine, lofepramine,
  nortriptyline, trimipramine
- mianserin, trazodone, phenelzine, isocarboxazid, tranylcypromine, moclobemide
- citalopram, escitalopram, fluoxetine, fluvoxamine, paroxetine, sertraline
- mirtazapine, reboxetine, venlafaxine, agomelatine, duloxetine
- flupentixol, tryptophan

Sorted, that is:

.. code-block:: none

    agomelatine
    amitriptyline
    citalopram
    clomipramine
    dosulepin
    doxepin
    duloxetine
    escitalopram
    fluoxetine
    flupentixol
    fluvoxamine
    imipramine
    isocarboxazid
    lofepramine
    mianserin
    mirtazapine
    moclobemide
    nortriptyline
    paroxetine
    phenelzine
    reboxetine
    sertraline
    tranylcypromine
    trazodone
    tricyclic
    trimipramine
    tryptophan
    venlafaxine

Compare that against the output of:

.. code-block:: python

    [x.generic_name for x in all_drugs_where(slam_antidepressant_finder=True,
                                             include_categories=True)]

**Using this code from R via reticulate**

Test within R:

.. code-block:: r

    # -------------------------------------------------------------------------
    # Load libraries
    # -------------------------------------------------------------------------

    RUN_ONCE_ONLY <- '
        library(devtools)
        devtools::install_github("rstudio/reticulate")  # get latest version
    '

    library(data.table)
    library(reticulate)

    # -------------------------------------------------------------------------
    # Set up reticulate
    # -------------------------------------------------------------------------

    VENV <- "~/dev/venvs/cardinal_pythonlib"  # or your preferred virtualenv
    PYTHON_EXECUTABLE <- ifelse(
        .Platform$OS.type == "windows",
        file.path(VENV, "Scripts", "python.exe"),  # Windows
        file.path(VENV, "bin", "python")  # Linux
    )
    reticulate::use_python(PYTHON_EXECUTABLE, required=TRUE)
    # ... it is CRITICAL to use required=TRUE, or it might fail silently

    # Unnecessary now reticulate::use_python() works:
    # 
    # PYTHON_VERSION <- "python3.5"
    # CARDINAL_PYTHONLIB_BASEDIR <- ifelse(
    #     .Platform$OS.type == "windows",
    #     file.path(VENV, "lib", "site-packages/cardinal_pythonlib"),
    #     file.path(VENV, "lib", PYTHON_VERSION, "site-packages/cardinal_pythonlib")
    # )
    # reticulate::use_virtualenv(VENV, required=TRUE)
    #
    # cpl_fileops <- reticulate::import_from_path("fileops", CARDINAL_PYTHONLIB_BASEDIR)
    # cpl_drugs <- reticulate::import_from_path("drugs", file.path(CARDINAL_PYTHONLIB_BASEDIR, "psychiatry"))
    #
    # ... this is NOT WORKING properly; dotted imports via reticulate::import() fail; also, imports from
    # within the Python code fail even if you use reticulate::import_from_path(); this suggests the virtualenv is not set up
    # properly; use reticulate::use_python() instead.

    # -------------------------------------------------------------------------
    # Import Python modules
    # -------------------------------------------------------------------------

    cardinal_pythonlib <- reticulate::import("cardinal_pythonlib")
    cpl_fileops <- reticulate::import("cardinal_pythonlib.fileops")
    cpl_drugs <- reticulate::import("cardinal_pythonlib.psychiatry.drugs")

    # -------------------------------------------------------------------------
    # Do something useful
    # -------------------------------------------------------------------------

    testnames <- c("citalopram", "Cipramil", "Prozac", "fluoxetine")
    # Works for simple variables:
    cpl_drugs$drug_names_to_generic(testnames)

    # Also works for data table replacements:
    dt <- data.table(
        subject = c("Alice", "Bob", "Charles", "Dawn", "Egbert", "Flora"),
        drug = c("citalopram", "Cipramil", "Prozac", "fluoxetine", "Priadel", "Haldol")
    )
    dt[, drug_generic := cpl_drugs$drug_names_to_generic(drug)]
    dt[, is_antidepressant := cpl_drugs$drug_names_match_criteria(
            drug_generic, 
            names_are_generic=TRUE,
            antidepressant=TRUE)]
    dt[, is_antidepressant_not_ssri := cpl_drugs$drug_names_match_criteria(
            drug_generic, 
            names_are_generic=TRUE,
            antidepressant=TRUE,
            ssri=FALSE)]
    dt[, is_conventional_antidepressant := cpl_drugs$drug_names_match_criteria(
            drug_generic, 
            names_are_generic=TRUE,
            conventional_antidepressant=TRUE)]
    dt[, slam_antidepressant_finder := cpl_drugs$drug_names_match_criteria(
            drug_generic, 
            names_are_generic=TRUE,
            slam_antidepressant_finder=TRUE,
            include_categories=TRUE)]

**Use for SQL finding**

.. code-block:: python

    from typing import List
    from cardinal_pythonlib.psychiatry.drugs import *
    
    colname = "somecol"
    
    antidepressants = all_drugs_where(conventional_antidepressant=True)  # type: List[Drug]
    antidep_sql_parts = [drug.sql_column_like_drug(colname) for drug in antidepressants]
    antidep_sql = " OR ".join(antidep_sql_parts)
    
    antipsychotics = all_drugs_where(antipsychotic=True)  # type: List[Drug]
    antipsy_sql_parts = [drug.sql_column_like_drug(colname) for drug in antipsychotics]
    antipsy_sql = " OR ".join(antipsy_sql_parts)
    
    alldrugs = all_drugs_where()
    alldrug_sql_parts = [drug.sql_column_like_drug(colname) for drug in alldrugs]
    alldrug_sql = " OR ".join(alldrug_sql_parts)
    
    lithium = get_drug("lithium")
    lithium_sql = lithium.sql_column_like_drug(colname)
    # HOWEVER, NOTE THAT LITHIUM IS CURRENTLY OVER-INCLUSIVE and will include
    # lithium chloride for LiDCO measurement.

"""  # noqa

import re
from typing import Dict, List, Optional, Pattern, Union

from cardinal_pythonlib.sql.literals import sql_string_literal


# =============================================================================
# Regex constants
# =============================================================================

WILDCARD = ".*"  # if re.DOTALL is set, this also matches newlines
WB = WORD_BOUNDARY = r"\b"


# =============================================================================
# Class to capture drug information
# =============================================================================

class Drug(object):
    """
    Class to describe a specific drug, or a drug category.

    Also embodies knowledge about brand names and common misspellings.

    See the :const:`DRUGS` list for example of use.
    """
    def __init__(
            self,
            # Names
            generic: Union[str, List[str]],
            alternatives: List[str] = None,
            category_not_drug: bool = False,
            add_preceding_wildcards: bool = True,
            add_preceding_word_boundary: bool = True,
            add_following_wildcards: bool = True,
            # Psychiatry
            psychotropic: bool = None,  # special; can be used as override if False  # noqa
            antidepressant: bool = False,
            conventional_antidepressant: bool = False,
            ssri: bool = False,
            non_ssri_modern_antidepressant: bool = False,
            tricyclic_antidepressant: bool = False,
            tetracyclic_and_related_antidepressant: bool = False,
            monoamine_oxidase_inhibitor: bool = False,
            antipsychotic: bool = False,
            first_generation_antipsychotic: bool = False,
            second_generation_antipsychotic: bool = False,
            stimulant: bool = False,
            anticholinergic: bool = False,
            benzodiazepine: bool = False,
            z_drug: bool = False,
            non_benzodiazepine_anxiolytic: bool = False,
            gaba_a_functional_agonist: bool = False,
            gaba_b_functional_agonist: bool = False,
            mood_stabilizer: bool = False,
            # Endocrinology
            antidiabetic: bool = False,
            sulfonylurea: bool = False,
            biguanide: bool = False,
            glifozin: bool = False,
            glp1_agonist: bool = False,
            dpp4_inhibitor: bool = False,
            meglitinide: bool = False,
            thiazolidinedione: bool = False,
            # Cardiovascular
            cardiovascular: bool = False,
            beta_blocker: bool = False,
            ace_inhibitor: bool = False,
            statin: bool = False,
            # Respiratory
            respiratory: bool = False,
            beta_agonist: bool = False,
            # Gastrointestinal
            gastrointestinal: bool = False,
            proton_pump_inhibitor: bool = False,
            nonsteroidal_anti_inflammatory: bool = False,
            # Nutritional
            vitamin: bool = False,
            # Special flags:
            slam_antidepressant_finder: bool = False) -> None:
        # noinspection PyUnresolvedReferences
        """
        Initialize and determine/store category knowledge.

        ``alternatives`` can include regexes (as text).

        We add front/back wildcards by default; this handles all situations
        like "depot X", etc. We also add a preceding word boundary (after the
        wildcard); thus the usual transformation is ``XXX`` -> ``.*\bXXX.*``.

        Args:
            generic: generic name, or list of names
            alternatives: can include regexes (as text)

            category_not_drug: is this a drug category, not a specific drug?

            add_preceding_wildcards: when making a regex (etc.), add a wildcard
                to the start of all possibilities (generic + alternative names)
                that don't already have one?
            add_preceding_word_boundary: when making a regex (etc.), add word
                boundaries to the start of all possibilities (generic +
                alternative names) that don't already have one?
            add_following_wildcards: when making a regex (etc.), add a wildcard
                to the end of all possibilities (generic + alternative names)
                that don't already have one?

            psychotropic: a psychotropic drug?

            antidepressant: an antidepressant?
            conventional_antidepressant: a traditional antidepressant?
            ssri: a selective serotonin reuptake inhibitor (SSRI)?
            non_ssri_modern_antidepressant: a non-SSRI "modern" antidepressant?
            tricyclic_antidepressant: a tricyclic?
            tetracyclic_and_related_antidepressant: a tetracyclic or related?
            monoamine_oxidase_inhibitor: a MAO-I?

            antipsychotic: an antipsychotic?
            first_generation_antipsychotic: an FGA?
            second_generation_antipsychotic: an SGA?

            stimulant: a psychostimulant?

            anticholinergic: an anticholinergic?

            benzodiazepine: a benzodiazepine?
            z_drug: a "Z" drug (e.g. zopiclone, zolpidem, ...)
            non_benzodiazepine_anxiolytic: a non-BZ anxiolytic?
            gaba_a_functional_agonist: a GABA-A functional agonist?
            gaba_b_functional_agonist: a GABA-B functional agonist?

            mood_stabilizer: a "mood stabilizer"?

            antidiabetic: treats diabetes?
            sulfonylurea: a sulfonylurea (sulphonylurea), for diabetes?
            biguanide: a biguanide, for diabetes?
            glifozin: a glifozin, for diabetes?
            glp1_agonist: a GLP-1 agonist, for diabetes?
            dpp4_inhibitor: a DPP4 inhibitor, for diabetes?
            meglitinide: a meglitinide, for diabetes?
            thiazolidinedione: a thiazolidinedione, for diabetes?

            cardiovascular: a cardiovascular drug?
            beta_blocker: a beta adrenoceptor antagonist?
            ace_inhibitor: an ACE inhibitor?
            statin: a statin?

            respiratory: a respiratory drug?
            beta_agonist: a beta adrenoceptor agonist?

            gastrointestinal: a gastrointestinal drug?
            proton_pump_inhibitor: a PPI?
            nonsteroidal_anti_inflammatory: an NSAID?

            vitamin: a vitamin?

            slam_antidepressant_finder: a drug found by the SLAM
                antidepressant-finding code? (A bit specialized, this one!)

        Attributes:

            mixture (bool): is this a mixture of more than one drug?
                Will be set if more than one generic name is given.
            all_generics (List[str]): list of all generic names in lower case
            generic_name: generic name (or combination name like ``a_with_b``
                for mixtures of ``a`` and ``b``)
            regex: compiled case-insensitive regular expression to match
                possible names
        """
        self.add_preceding_word_boundary = add_preceding_word_boundary
        self.add_preceding_wildcards = add_preceding_wildcards
        self.add_following_wildcards = add_following_wildcards

        # ---------------------------------------------------------------------
        # Name handling
        # ---------------------------------------------------------------------

        if isinstance(generic, list):
            self.mixture = True
            self.all_generics = [x.lower().strip() for x in generic]
            self.generic_name = "_with_".join(self.all_generics)
        elif isinstance(generic, str):
            self.mixture = False
            self.generic_name = generic.lower().strip()
            self.all_generics = [self.generic_name]
        else:
            raise ValueError(f"Bad generic_name: {generic!r}")
        self.alternatives = alternatives or []  # type: List[str]
        self._regex_text = None  # type: Optional[str]
        self._regex = None  # type: Optional[Pattern]
        self._sql_like_fragments = None  # type: Optional[List[str]]

        # ---------------------------------------------------------------------
        # Things we know about psychotropics
        # ---------------------------------------------------------------------

        if (ssri or non_ssri_modern_antidepressant or
                tricyclic_antidepressant or
                tetracyclic_and_related_antidepressant or
                monoamine_oxidase_inhibitor):
            conventional_antidepressant = True

        if conventional_antidepressant:
            antidepressant = True

        if first_generation_antipsychotic or second_generation_antipsychotic:
            antipsychotic = True

        if benzodiazepine or z_drug:
            gaba_a_functional_agonist = True

        if ((antidepressant or antipsychotic or stimulant or anticholinergic or
                gaba_a_functional_agonist or gaba_b_functional_agonist or
                mood_stabilizer) and
                (psychotropic is not False)):
            psychotropic = True
        if psychotropic is None:
            psychotropic = False

        # ---------------------------------------------------------------------
        # Things we know about other drugs
        # ---------------------------------------------------------------------

        if (sulfonylurea or biguanide or glifozin or glp1_agonist or
                dpp4_inhibitor or meglitinide or thiazolidinedione):
            antidiabetic = True

        if beta_blocker or ace_inhibitor:
            cardiovascular = True

        # ---------------------------------------------------------------------
        # Store category knowledge
        # ---------------------------------------------------------------------

        self.category_not_drug = category_not_drug

        self.psychotropic = psychotropic
        self.antidepressant = antidepressant
        self.conventional_antidepressant = conventional_antidepressant
        self.ssri = ssri
        self.non_ssri_modern_antidepressant = non_ssri_modern_antidepressant
        self.tricyclic = tricyclic_antidepressant
        self.tetracyclic_and_related_antidepressant = tetracyclic_and_related_antidepressant  # noqa
        self.monoamine_oxidase_inhibitor = monoamine_oxidase_inhibitor

        self.antipsychotic = antipsychotic
        self.first_generation_antipsychotic = first_generation_antipsychotic
        self.second_generation_antipsychotic = second_generation_antipsychotic

        self.stimulant = stimulant

        self.anticholinergic = anticholinergic

        self.benzodiazepine = benzodiazepine
        self.z_drug = z_drug
        self.gaba_a_functional_agonist = gaba_a_functional_agonist
        self.gaba_b_functional_agonist = gaba_b_functional_agonist
        self.non_benzodiazepine_anxiolytic = non_benzodiazepine_anxiolytic

        self.mood_stabilizer = mood_stabilizer

        self.antidiabetic = antidiabetic
        self.sulfonylurea = sulfonylurea
        self.biguanide = biguanide
        self.cardiovascular = cardiovascular
        self.beta_blocker = beta_blocker
        self.ace_inhibitor = ace_inhibitor
        self.statin = statin
        self.respiratory = respiratory
        self.beta_agonist = beta_agonist
        self.gastrointestinal = gastrointestinal
        self.proton_pump_inhibitor = proton_pump_inhibitor
        self.nonsteroidal_anti_inflammatory = nonsteroidal_anti_inflammatory
        self.vitamin = vitamin

        # ---------------------------------------------------------------------
        # Store other flags
        # ---------------------------------------------------------------------

        self.slam_antidepressant_finder = slam_antidepressant_finder

    @property
    def regex_text(self) -> str:
        """
        Return regex text (yet to be compiled) for this drug.
        """
        if self._regex_text is None:
            possibilities = []  # type: List[str]
            for p in list(set(self.all_generics + self.alternatives)):
                if self.add_preceding_word_boundary and not p.startswith(WB):
                    p = WB + p
                if self.add_preceding_wildcards and not p.startswith(WILDCARD):
                    p = WILDCARD + p
                if self.add_following_wildcards and not p.endswith(WILDCARD):
                    p = p + WILDCARD
                possibilities.append(p)
            self._regex_text = "|".join("(?:" + x + ")" for x in possibilities)
        return self._regex_text

    @property
    def regex(self) -> Pattern:
        """
        Returns a compiled regex for this drug.
        """
        if self._regex is None:
            self._regex = re.compile(self.regex_text,
                                     re.IGNORECASE | re.DOTALL)
        return self._regex

    @staticmethod
    def regex_to_sql_like(regex_text: str,
                          single_wildcard: str = "_",
                          zero_or_more_wildcard: str = "%") -> List[str]:
        """
        Converts regular expression text to a reasonably close fragment
        for the SQL ``LIKE`` operator.

        NOT PERFECT, but works for current built-in regular expressions.

        Args:
            regex_text: regular expression text to work with
            single_wildcard: SQL single wildcard, typically an underscore
            zero_or_more_wildcard: SQL "zero/one/many" wildcard, probably always
                a percent symbol

        Returns:
            string for an SQL string literal

        Raises:
            :exc:`ValueError` for some regex text that it doesn't understand
            properly
        """
        def append_to_all(new_content: str) -> None:
            nonlocal results
            results = [r + new_content for r in results]

        def split_and_append(new_options: List[str]) -> None:
            nonlocal results
            newresults = []  # type: List[str]
            for option in new_options:
                newresults.extend([r + option for r in results])
            results = newresults

        def deduplicate_wildcards(text: str) -> str:
            while zero_or_more_wildcard + zero_or_more_wildcard in text:
                text = text.replace(
                    zero_or_more_wildcard + zero_or_more_wildcard,
                    zero_or_more_wildcard)
            return text

        # Basic processing
        working = regex_text  # strings are immutable
        results = [zero_or_more_wildcard]  # start with a wildcard

        while working:
            if working.startswith(".*"):
                # e.g. ".*ozapi"
                append_to_all(zero_or_more_wildcard)
                working = working[2:]
            elif working.startswith("["):
                # e.g. "[io]peridol"
                close_bracket = working.index("]")  # may raise
                bracketed = working[1:close_bracket]
                option_groups = bracketed.split("|")
                options = [c for group in option_groups for c in group]
                split_and_append(options)
                working = working[close_bracket + 1:]
            elif len(working) > 1 and working[1] == "?":
                # e.g. "r?azole"
                split_and_append(["", working[0]])
                # ... regex "optional character"
                # ... SQL: some results with a single wildcard, some without
                working = working[2:]
            elif working.startswith("."):
                # single character wildcard
                append_to_all(single_wildcard)
                working = working[1:]
            else:
                append_to_all(working[0])
                working = working[1:]
        append_to_all(zero_or_more_wildcard)  # end with a wildcard

        # Remove any duplicate (consecutive) % wildcards:
        results = [deduplicate_wildcards(r) for r in results]

        # Done
        return results

    @property
    def sql_like_fragments(self) -> List[str]:
        """
        Returns all the string literals to which a database column should be
        compared using the SQL ``LIKE`` operator, to match this drug.

        This isn't as accurate as the regex, but ``LIKE`` can do less.

        ``LIKE`` uses the wildcards ``?`` and ``%``.
        """
        if self._sql_like_fragments is None:
            self._sql_like_fragments = []
            for p in list(set(self.all_generics + self.alternatives)):
                self._sql_like_fragments.extend(self.regex_to_sql_like(p))
        return self._sql_like_fragments

    def name_matches(self, name: str) -> bool:
        """
        Detects whether the name that's passed matches our knowledge of any of
        things that this drug might be called: generic name, brand name(s),
        common misspellings.

        The parameter should be pre-stripped of edge whitespace.
        """
        return bool(self.regex.match(name))

    def sql_column_like_drug(self, column_name: str) -> str:
        """
        Returns SQL like

        .. code-block:: sql

            (column_name LIKE '%drugname1%' OR
             column_name LIKE '%drugname2%')

        for the drug names that this Drug object knows about.

        Args:
            column_name: column name, pre-escaped if necessary

        Returns:
            SQL fragment as above

        """
        clauses = [
            f"{column_name} LIKE {sql_string_literal(f)}"
            for f in self.sql_like_fragments
        ]
        return f"({' OR '.join(clauses)})"


# Source data.
DRUGS = [
    # In comments below: (*) misspelling, capitalized for brand name, (~)
    # hybrid generic/brand name, (+) old name.

    # -------------------------------------------------------------------------
    # SSRIs
    # -------------------------------------------------------------------------
    Drug(
        "citalopram",
        ["Cipramil", "Celexa"],
        ssri=True,
        slam_antidepressant_finder=True
    ),
    Drug(
        "escitalopram",
        ["Cipralex", "Lexapro"],
        ssri=True,
        slam_antidepressant_finder=True
    ),
    Drug(
        "fluoxetine",
        ["Prozac", "Bellzac", "Oxactin", "Prozep", "Sarafem", "fluox.*"],
        # CPFT 2013: "fluoxetine  Dec"
        ssri=True,
        slam_antidepressant_finder=True
    ),
    Drug(
        "fluvoxamine",
        ["Luvox", "Faverin", "fluvoxamine.*"],  # e.g. "fluvoxamine maleate"
        ssri=True,
        slam_antidepressant_finder=True
    ),
    Drug(
        "paroxetine",
        ["Seroxat", "Paxil"],  # there are other brands elsewhere...
        ssri=True,
        slam_antidepressant_finder=True
    ),
    Drug(
        "sertraline",
        ["Lustral", "Zoloft", "Bellsert"],
        # NOT Seretra (cf. SLAM code, see email to self 2016-12-02); Seretra =
        # seratrodast = for asthma
        ssri=True,
        slam_antidepressant_finder=True
    ),

    # -------------------------------------------------------------------------
    # FIRST-GENERATION ANTIPSYCHOTICS
    # -------------------------------------------------------------------------
    Drug("benperidol", ["Anquil"], first_generation_antipsychotic=True),
    Drug("chlorpromazine", ["Largactil"], first_generation_antipsychotic=True),
    Drug(
        "flupentixol",
        ["Depixol", "Fluanxol", "flupent.*", "Depixol.*"],
        # e.g. flupenthixol, flupenthixol decanoate, flupentixol decanoate
        first_generation_antipsychotic=True,
        antidepressant=True,
        slam_antidepressant_finder=True
    ),
    Drug(
        "fluphenazine",
        ["Modecate", "fluphen.*", "Modecate.*"],
        first_generation_antipsychotic=True
    ),
    Drug(
        "haloperidol",
        [
            "Haldol", "Serenase",
            "hal[io]p.*", "Dozi.*", "Hald.*", "Serena.*",
            # NB Serenase, Serenace.
            #  CPFT 2013: haloperidol, haloperidol decanoate, Haldol, Haldol
            #  decanoate, Serenase.
        ],
        first_generation_antipsychotic=True
    ),
    Drug("levomepromazine", ["Nozinan"], first_generation_antipsychotic=True),
    Drug("pericyazine", first_generation_antipsychotic=True),
    Drug("perphenazine", ["Fentazin"], first_generation_antipsychotic=True),
    Drug(
        ["amitriptyline", "perphenazine"],
        ["Triptafen"],  # special
        tricyclic_antidepressant=True,
        slam_antidepressant_finder=True
    ),
    Drug("pimozide", ["Orap"], first_generation_antipsychotic=True),
    Drug(
        "pipotiazine",
        ["pipot.*", "Piport.*"],
        # ... actually (CPFT 2013): pipotiazine, Piportil
        first_generation_antipsychotic=True
    ),
    Drug(
        "prochlorperazine",
        ["Stemetil"],
        first_generation_antipsychotic=True
    ),
    Drug("promazine", first_generation_antipsychotic=True),
    Drug(
        "sulpiride",
        ["Dolmatil", "Sulpor"],
        first_generation_antipsychotic=True
    ),
    Drug(
        "trifluoperazine",
        ["Stelazine"],
        first_generation_antipsychotic=True
    ),
    Drug(
        "zuclopenthixol",
        ["zuclop.*", "Clopix.*", "Acc?uphase"],
        # ... actually (CPFT 2013): zuclopenthixol, zuclopenthixol acetate,
        # zuclopenthixol decanoate, Clopixol, Clopixol Decanoate, Acuphase
        first_generation_antipsychotic=True
    ),

    # -------------------------------------------------------------------------
    # SECOND-GENERATION ANTIPSYCHOTICS
    # -------------------------------------------------------------------------
    Drug(
        "amisulpride",
        ["amisulp.*", "Solian"],
        # ... actually (CPFT 2013): amisulpiride(*), amisulpride, Solian
        second_generation_antipsychotic=True
    ),
    Drug(
        "aripiprazole",
        ["Abilify", "ari?pr?ipr?azol.*"],
        second_generation_antipsychotic=True
    ),
    Drug(
        "asenapine",
        ["Saphris", "Sycrest"],
         second_generation_antipsychotic=True
    ),
    Drug(
        "clozapine",
        ["cloz.*", "Denz.*", "Zapon.*"],
        # ... actually (CPFT 2013): clozapine, Clozaril, clozepine(*)
        second_generation_antipsychotic=True
    ),
    Drug(
        "iloperidone",
        ["Fanapt", "Fanapta", "Zomaril"],
        second_generation_antipsychotic=True
    ),
    Drug("lurasidone", ["Latuda"], second_generation_antipsychotic=True),
    Drug(
        "olanzapine",
        ["olanz.*", "Zalast.*", "Zyprex.*", "Zypad.*"],
        # ... actually (CPFT 2013): olanzapine, olanzapine embonate,
        # olanz(*), olanzepine(*), olanzapin(*), Zyprexa
        second_generation_antipsychotic=True
    ),
    Drug(
        "paliperidone",
        ["Invega", "Xeplion"],
        second_generation_antipsychotic=True
    ),
    Drug(
        "quetiapine",
        ["quet.*", "Seroquel"],
        # ... actually (CPFT 2013): quetiapine, quetiepine(*), Seroquel
        second_generation_antipsychotic=True
    ),
    Drug(
        "risperidone",
        ["risp.*", "Consta"],
        # ... actually (CPFT 2013): risperidone, risperadone(*), Risperidone
        # Consta (~), Risperdal, Risperdal Consta
        second_generation_antipsychotic=True
    ),
    Drug(
        "sertindole",
        ["Serdolect", "Serlect"],
        second_generation_antipsychotic=True
    ),
    Drug("ziprasidone", second_generation_antipsychotic=True),
    Drug(
        "zotepine",  # not in UK
        ["Nipolept", "Losizopilon", "Lodopin", "Setous"],
        second_generation_antipsychotic=True
    ),

    # -------------------------------------------------------------------------
    # STIMULANTS
    # -------------------------------------------------------------------------
    Drug(
        "amfetamine",
        [".*am[ph|f]etamine.*", "Adderall"],
        # ... actually (CPFT 2013): dextroamphetamine(+), dexamfetamine
        stimulant=True
    ),
    Drug(
        "methylphenidate",
        ["Ritalin", "Concerta.*", "Equasym.*", "Medikinet.*"],
        # ... actually (CPFT 2013): methylphenidate, Ritalin, Concerta
        stimulant=True
    ),
    Drug("modafinil", ["Provigil"], stimulant=True),

    # -------------------------------------------------------------------------
    # ANTICHOLINERGICS
    # -------------------------------------------------------------------------
    Drug("benztropine", ["benzatropine"], anticholinergic=True),
    Drug("orphenadrine", ["Biorphen", "Disipal"], anticholinergic=True),
    Drug("procyclidine", ["Arpicolin", "Kemadrin"], anticholinergic=True),
    Drug("trihexyphenidyl", ["Broflex"], anticholinergic=True),

    # -------------------------------------------------------------------------
    # OTHER MODERN ANTIDEPRESSANTS
    # -------------------------------------------------------------------------
    Drug(
        "agomelatine",
        ["Valdoxan"],
        non_ssri_modern_antidepressant=True,
        slam_antidepressant_finder=True
    ),
    Drug(
        "bupropion",
        ["Zyban"],
        non_ssri_modern_antidepressant=True
        # antidepressant license in US, smoking cessation in UK
    ),
    Drug(
        "duloxetine",
        ["Cymbalta", "Yentreve", "duloxat.*"],
         non_ssri_modern_antidepressant=True,
        slam_antidepressant_finder=True
    ),
    Drug(
        "mirtazapine",
        ["mirtaz.*", "mirtazepine", "Zispin", "Mirza"],
        # ... actually (CPFT 2013): mirtazapine, mirtazepine(*), "mirtazapine
        # Dec" (?)
        non_ssri_modern_antidepressant=True,
        slam_antidepressant_finder=True
    ),
    Drug(
        "reboxetine",
        ["Edronax", "reboxat.*"],
        non_ssri_modern_antidepressant=True,
        slam_antidepressant_finder=True
    ),
    Drug(
        "tryptophan",
        ["Optimax"],
        non_ssri_modern_antidepressant=True,
        slam_antidepressant_finder=True
    ),
    Drug(
        "venlafaxine",
        ["venla.*", "Eff?exor.*"],
        # ... actually (CPFT 2013): venlafaxine, venlafaxine XL,
        non_ssri_modern_antidepressant=True,  # though obviously an SSRI too...
        slam_antidepressant_finder=True
    ),

    # -------------------------------------------------------------------------
    # TRICYCLIC AND RELATED ANTIDEPRESSANTS
    # -------------------------------------------------------------------------
    Drug(
        "tricyclic_antidepressant",
        ["tricyclic.*", "tca" + WB],
        tricyclic_antidepressant=True,
        slam_antidepressant_finder=True,
        category_not_drug=True,
    ),
    Drug(
        "amitriptyline",
        ["amitr[i|y]pt[i|y]l[i|y]n.*", "Vanatrip", "Elavil", "Endep"],
        # ... actually (CPFT 2013): amitriptyline, amitriptiline(*),
        # amitryptyline(*)
        # Triptafen = amitriptyline + perphenazine; see above.
        tricyclic_antidepressant=True,
        slam_antidepressant_finder=True
    ),
    Drug(
        "clomipramine",
        ["Anafranil.*"],
        tricyclic_antidepressant=True,
        slam_antidepressant_finder=True
    ),
    Drug(
        "dosulepin",
        ["dothiepin", "Prothiaden"],
        # ... actually (CPFT 2013): dosulepin, dothiepin(+)
        tricyclic_antidepressant=True,
        slam_antidepressant_finder=True
    ),
    Drug(
        "doxepin",
        ["Sinepin", "Sinequan", "Sinepin", "Xepin"],
        # Xepin is cream only
        tricyclic_antidepressant=True,
        slam_antidepressant_finder=True
    ),
    Drug(
        "imipramine",
        ["Tofranil"],
        tricyclic_antidepressant=True,
        slam_antidepressant_finder=True
    ),
    Drug(
        "lofepramine",
        ["Lomont"],
        tricyclic_antidepressant=True,
        slam_antidepressant_finder=True
    ),
    Drug(
        "nortriptyline",
        ["nortr.*", "Allegron", "Pamelor", "Aventyl"],
        # ... actually (CPFT 2013): nortriptyline, nortryptiline(*)
        tricyclic_antidepressant=True,
        slam_antidepressant_finder=True
    ),
    Drug(
        "trimipramine",
        ["Surmontil"],
        tricyclic_antidepressant=True,
        slam_antidepressant_finder=True
    ),

    # -------------------------------------------------------------------------
    # TETRACYCLIC-RELATED ANTIDEPRESSANTS
    # -------------------------------------------------------------------------
    Drug(
        "mianserin",
        tetracyclic_and_related_antidepressant=True,
        slam_antidepressant_finder=True
    ),
    Drug(
        "trazodone",
        ["Molipaxin"],
        tetracyclic_and_related_antidepressant=True,
        slam_antidepressant_finder=True
    ),
    Drug(
        "nefazodone",
        # discontinued for hepatotoxicity? But apparently still used in 2014
        # in the UK: http://www.bbc.co.uk/news/uk-25745824
        ["Dutonin", "Nefadar", "Serzone"],
        tetracyclic_and_related_antidepressant=True
        # brand names from https://en.wikipedia.org/wiki/Nefazodone
        # ... yup, still a trickle, mostly from Islington:
        # https://openprescribing.net/chemical/0403040T0/
    ),
    Drug(
        "maprotiline",
        ["Ludiomil"],
        tetracyclic_and_related_antidepressant=True
    ),

    # -------------------------------------------------------------------------
    # MAOIs
    # -------------------------------------------------------------------------
    Drug(
        "phenelzine",
        ["phenylethylhydrazine", "Alazin", "Nardil"],
        monoamine_oxidase_inhibitor=True,
        slam_antidepressant_finder=True
        # - SLAM code (see e-mail to self 2016-12-02) also has %Alazin%; not sure  # noqa
        #   that's right; see also
        #   http://www.druglib.com/activeingredient/phenelzine/
        # - oh, yes, it is right:
        #   https://www.pharmacompass.com/active-pharmaceutical-ingredients/alazin  # noqa
        # - phenylethylhydrazine is a synonym; see
        #   http://www.minclinic.ru/drugs/drugs_eng/B/Beta-phenylethylhydrazine.html  # noqa
    ),
    # not included: pheniprazine
    Drug(
        "isocarboxazid",
        monoamine_oxidase_inhibitor=True,
        slam_antidepressant_finder=True
    ),
    Drug(
        "moclobemide",
        ["Manerix"],
        monoamine_oxidase_inhibitor=True,
        slam_antidepressant_finder=True
    ),
    Drug(
        "tranylcypromine",
        ["Parnate"],
        monoamine_oxidase_inhibitor=True,
        slam_antidepressant_finder=True
    ),

    # -------------------------------------------------------------------------
    # BENZODIAZEPINES
    # -------------------------------------------------------------------------
    Drug(
        "benzodiazepine",
        ["benzodiazepine.*"],
        benzodiazepine=True,
        category_not_drug=True
    ),
    Drug("alprazolam", benzodiazepine=True),
    Drug("chlordiazepoxide", benzodiazepine=True),
    Drug("clobazam", benzodiazepine=True),
    Drug("clonazepam", ["Rivotril"], benzodiazepine=True),
    Drug(
        "diazepam",
        ["diaz.*", "Valium"],
        # ... actually (CPFT 2013): diazepam, diazapam(*), diazapem(*), Valium
        benzodiazepine=True
    ),
    Drug("flurazepam", ["Dalmane"], benzodiazepine=True),
    Drug("loprazolam", benzodiazepine=True),
    Drug("lorazepam", ["Ativan"], benzodiazepine=True),
    Drug("lormetazepam", benzodiazepine=True),
    Drug("midazolam", ["Hypnovel"], benzodiazepine=True),
    Drug("nitrazepam", benzodiazepine=True),
    Drug("oxazepam", benzodiazepine=True),
    Drug("temazepam", benzodiazepine=True),

    # -------------------------------------------------------------------------
    # Z-DRUGS
    # -------------------------------------------------------------------------
    Drug("zaleplon", ["Sonata"], z_drug=True),
    Drug(
        "zolpidem",
        ["zolpidem.*", "Stilnoct"],
        # ... actually (CPFT 2013): zolpidem, zolpidem tartrate
        z_drug=True
    ),
    Drug("zopiclone", ["Zimovane"], z_drug=True),

    # -------------------------------------------------------------------------
    # OTHER GABA MODULATORS
    # -------------------------------------------------------------------------
    Drug(
        "baclofen",
        [
            "Lioresal", "Lyflex", "Bacfen", "Baclof", "Bacmax", "Chinofen",
            "Parafon", "Riclofen", "Spinofen", "Spinospas", "Tefsole",
            "Gablofen", "Kemstro"
        ],
        gaba_b_functional_agonist=True
    ),

    # -------------------------------------------------------------------------
    # OTHER ANXIOLYTICS
    # -------------------------------------------------------------------------
    Drug("buspirone", ["Buspar"], non_benzodiazepine_anxiolytic=True),

    # -------------------------------------------------------------------------
    # OTHER ANTIMANIC
    # -------------------------------------------------------------------------
    Drug(
        "carbamazepine",
        ["Carbagen.*", "Tegretol.*"],
        # also Tegretol Prolonged Release (formerly Tegretol Retard)
        # ... actually (CPFT 2013): carbamazepine, Tegretol
        mood_stabilizer=True
    ),
    Drug(
        "valproate",
        [".*valp.*", "Epilim.*", "Episenta", "Epival", "Convulex", "Depakote"],
        # ... also semisodium valproate
        # ... actually (CPFT 2013): sodium valproate [chrono], valproic acid,
        # valproate, sodium valproate, sodium valporate(*), sodium valporate(*)
        # chrono, Depakote
        mood_stabilizer=True
    ),
    Drug(
        "lithium",
        ["lithium.*", "Camcolit", "Liskonum", "Priadel", "Li-Liquid"],
        # ... actually (CPFT 2013): lithium, lithium carbonate, lithium citrate
        # (curious: Priadel must be being changed to lithium...)
        antidepressant=True,
        mood_stabilizer=True
    ),

    # -------------------------------------------------------------------------
    # OTHER FOR BIPOLAR/UNIPOLAR DEPRESSION
    # -------------------------------------------------------------------------
    Drug(
        "lamotrigine",
        ["lamotrigine.*", "Lamictal"],
        mood_stabilizer=True,
        antidepressant=True,
    ),
    Drug(
        "triiodothyronine",
        ["tri-iodothyronine", "liothyronine", "Cytomel"],
        antidepressant=True,
    ),

    # -------------------------------------------------------------------------
    # GENERAL MEDICINE: DIABETES
    # -------------------------------------------------------------------------
    Drug("glibenclamide", sulfonylurea=True),
    Drug(
        "gliclazide",
        ["Zicron", "Diamicron.*", "Dacadis.*", "Vitile.*"],
        sulfonylurea=True
    ),
    Drug("glimepiride", ["Amaryl"], sulfonylurea=True),
    Drug("glipizide", ["Minodiab"], sulfonylurea=True),
    Drug("tolbutamide", sulfonylurea=True),
    Drug("metformin", ["metformin.*", "Glucophage.*"], biguanide=True),
    Drug("acarbose", ["Glucobay"], antidiabetic=True),
    Drug("dapagliflozin", ["Forxiga"], glifozin=True),
    Drug("exenatide", ["Byetta", "Bydureon"], glp1_agonist=True),
    Drug("linagliptin", ["Trajenta"], dpp4_inhibitor=True),
    Drug(["linagliptin", "metformin"], ["Jentadueto"],
         biguanide=True, dpp4_inhibitor=True),
    Drug("liraglutide", ["Victoza"], glp1_agonist=True),
    Drug("lixisenatide", ["Lyxumia"], glp1_agonist=True),
    Drug("nateglinide", ["Starlix"], meglitinide=True),
    Drug("pioglitazone", ["Actos"], thiazolidinedione=True),
    Drug(["pioglitazone", "metformin"], ["Competact"],
         thiazolidinedione=True, biguanide=True),
    Drug("repaglinide", ["Prandin"], meglitinide=True),
    Drug("saxagliptin", ["Onglyza"], dpp4_inhibitor=True),
    Drug(["saxagliptin", "metformin"], ["Komboglyze"],
         dpp4_inhibitor=True, biguanide=True),
    Drug("sitagliptin", ["Januvia"], dpp4_inhibitor=True),
    Drug(["sitagliptin", "metformin"], ["Janumet"],
         dpp4_inhibitor=True, biguanide=True),
    Drug("vildagliptin", ["Galvus"], dpp4_inhibitor=True),
    Drug(["vildagliptin", "metformin"], ["Eucreas"], 
         dpp4_inhibitor=True, biguanide=True),
    Drug(
        "insulin",
        # Insulin. Covering the BNF categories:
        # INSULIN
        # INSULIN ASPART
        # INSULIN GLULISINE
        # INSULIN LISPRO
        # INSULIN DEGLUDEC
        # INSULIN DETEMIR
        # INSULIN GLARGINE
        # INSULIN ZINC SUSPENSION
        # ISOPHANE INSULIN
        # PROTAMINE ZINC INSULIN
        # BIPHASIC INSULIN ASPART
        # BIPHASIC INSULIN LISPRO
        # BIPHASIC ISOPHANE INSULIN
        [
            ".*insulin.*", ".*aspart.*", ".*glulisine.*", ".*lispro.*",
            ".*degludec.*", ".*detemir.*", ".*glargine.*", ".*Hypurin.*",
            ".*Actrapid.*", ".*Humulin.*", ".*Insuman.*", ".*Novorapid.*",
            ".*Apidra.*", ".*Humalog.*", ".*Tresiba.*", ".*Levemir.*",
            ".*Lantus.*", ".*Insulatard.*", ".*NovoMix.*",
        ], 
        antidiabetic=True
    ),

    # -------------------------------------------------------------------------
    # GENERAL MEDICINE: CARDIOVASCULAR
    # -------------------------------------------------------------------------
    Drug("aspirin", cardiovascular=True),
    Drug("atenolol", beta_blocker=True),
    # ACE inhibitors (selected)
    Drug("lisinopril", ace_inhibitor=True),
    Drug("ramipril", ace_inhibitor=True),
    # Statins
    Drug("atorvastatin", ["Lipitor"], statin=True),
    Drug("fluvastatin", ["Lescol.*"], statin=True),
    Drug("pravastatin", ["Lipostat"], statin=True),
    Drug("rosuvastatin", ["Crestor"], statin=True),
    Drug("simvastatin", ["Zocor"], statin=True),
    Drug(["simvastatin", "ezetimibe"], ["Inegy"], statin=True),

    # -------------------------------------------------------------------------
    # GENERAL MEDICINE: RESPIRATORY
    # -------------------------------------------------------------------------
    Drug(
        "salbutamol",
        ["salbut.*", "vent.*"],
        # ... actually (CPFT 2013): salbutamol
        respiratory=True, beta_agonist=True
    ),

    # -------------------------------------------------------------------------
    # GENERAL MEDICINE: GASTROINTESTINAL
    # -------------------------------------------------------------------------
    Drug(
        "lactulose",
        ["lactul.*", "Duphal.*", "Lactug.*", "laevol.*"],
        # ... actually (CPFT 2013): lactulose
        gastrointestinal=True
    ),
    Drug("lansoprazole", proton_pump_inhibitor=True),
    Drug("omeprazole", proton_pump_inhibitor=True),
    Drug("senna", gastrointestinal=True),

    # -------------------------------------------------------------------------
    # GENERAL MEDICINE: OTHER
    # -------------------------------------------------------------------------
    Drug("ibuprofen", nonsteroidal_anti_inflammatory=True),
    Drug("levothyroxine"),
    Drug("paracetamol"),
    Drug("thiamine", vitamin=True),

    # -------------------------------------------------------------------------
    # MAYBE ADD:
    # - OPIOIDS
    # - clonidine
    # - cloral betaine
    # - ?domperidone
    # - donepezil
    # - gabapentin
    # - hyoscine
    # - Keppra = levetiracetam
    # - linezolid (as it's an MAOI)
    # - memantine
    # - methyldopa
    # - ?metoclopramide
    # - nicotine
    # - pregabalin
    # - promethazine
    # - ropinirole
    # - rotigotine
    # - selegiline
    # - topiramate
    # -------------------------------------------------------------------------

]  # type: List[Drug]


# =============================================================================
# High-speed lookup versions of the original constants
# =============================================================================

DRUGS_BY_GENERIC_NAME = {d.generic_name: d for d in DRUGS}


# =============================================================================
# Get drug object by name
# =============================================================================

def get_drug(drug_name: str,
             name_is_generic: bool = False,
             include_categories: bool = False) -> Optional[Drug]:
    """
    Converts a drug name to a :class:`.Drug` class.

    If you already have the generic name, you can get the Drug more
    efficiently by setting ``name_is_generic=True``.

    Set ``include_categories=True`` to include drug categories (such as
    tricyclics) as well as individual drugs.
    """
    drug_name = drug_name.strip().lower()
    if name_is_generic:
        drug = DRUGS_BY_GENERIC_NAME.get(drug_name)  # type: Optional[Drug]
        if drug is not None and drug.category_not_drug and not include_categories:  # noqa
            return None
        return drug
    else:
        for d in DRUGS:
            if d.name_matches(drug_name):
                return d
        return None


# =============================================================================
# Convert drug names to generic equivalents
# =============================================================================

def drug_name_to_generic(drug_name: str,
                         unknown_to_default: bool = False,
                         default: str = None,
                         include_categories: bool = False) -> str:
    """
    Converts a drug name to the name of its generic equivalent.
    """
    drug = get_drug(drug_name, include_categories=include_categories)
    if drug is not None:
        return drug.generic_name
    return default if unknown_to_default else drug_name


def drug_names_to_generic(drugs: List[str],
                          unknown_to_default: bool = False,
                          default: str = None,
                          include_categories: bool = False) -> List[str]:
    """
    Converts a list of drug names to their generic equivalents.

    The arguments are as for :func:`drug_name_to_generic` but this function
    handles a list of drug names rather than a single one.

    Note in passing the following conversion of blank-type representations from
    R via ``reticulate``, when using e.g. the ``default`` parameter and storing
    results in a ``data.table()`` character column:

    .. code-block:: none

        ------------------------------  ----------------
        To Python                       Back from Python
        ------------------------------  ----------------
        [not passed, so Python None]    "NULL"
        NULL                            "NULL"
        NA_character_                   "NA"
        NA                              TRUE (logical)
        ------------------------------  ----------------

    """
    return [
        drug_name_to_generic(drug,
                             unknown_to_default=unknown_to_default,
                             default=default,
                             include_categories=include_categories)
        for drug in drugs
    ]


# =============================================================================
# Check drugs against criteria
# =============================================================================

def drug_matches_criteria(drug: Drug, **criteria: Dict[str, bool]) -> bool:
    """
    Determines whether a drug, passed as an instance of :class:`.Drug`, matches
    the specified criteria.

    Args:
        drug: a :class:`.Drug` instance
        criteria: ``name=value`` pairs to match against the attributes of
            the :class:`Drug` class. For example, you can include keyword
            arguments like ``antidepressant=True``.
    """
    for attribute, value in criteria.items():
        if getattr(drug, attribute) != value:
            return False
    return True


def all_drugs_where(sort=True,
                    include_categories: bool = False,
                    **criteria: Dict[str, bool]) -> List[Drug]:
    """
    Find all drugs matching the specified criteria (see
    :func:`drug_matches_criteria`). If ``include_categories`` is true, then
    drug categories (like "tricyclics") are included as well as individual
    drugs.

    Pass keyword arguments such as

    .. code-block:: python

        from cardinal_pythonlib.psychiatry.drugs import *
        non_ssri_antidep = all_drugs_where(antidepressant=True, ssri=False)
        print([d.generic_name for d in non_ssri_antidep])
        conventional_antidep = all_drugs_where(conventional_antidepressant=True)
        print([d.generic_name for d in conventional_antidep])
    """
    matching_drugs = []  # type: List[Drug]
    for drug in DRUGS:
        if drug.category_not_drug and not include_categories:
            continue
        if drug_matches_criteria(drug, **criteria):
            matching_drugs.append(drug)
    if sort:
        matching_drugs.sort(key=lambda d: d.generic_name)
    return matching_drugs


def drug_name_matches_criteria(drug_name: str,
                               name_is_generic: bool = False,
                               include_categories: bool = False,
                               **criteria: Dict[str, bool]) -> bool:
    """
    Establish whether a single drug, passed by name, matches the specified
    criteria. See :func:`drug_matches_criteria`.
    """
    drug = get_drug(drug_name, name_is_generic)
    if drug is None:
        return False
    if drug.category_not_drug and not include_categories:
        return False
    return drug_matches_criteria(drug, **criteria)


def drug_names_match_criteria(drug_names: List[str],
                              names_are_generic: bool = False,
                              include_categories: bool = False,
                              **criteria: Dict[str, bool]) -> List[bool]:
    """
    Establish whether multiple drugs, passed as a list of drug names, each
    matches the specified criteria. See :func:`drug_matches_criteria`.
    """
    return [
        drug_name_matches_criteria(
            dn,
            name_is_generic=names_are_generic,
            include_categories=include_categories,
            **criteria)
        for dn in drug_names
    ]



#!/usr/bin/env python
# cardinal_pythonlib/docker.py

"""
===============================================================================

    Original code copyright (C) 2009-2021 Rudolf Cardinal (rudolf@pobox.com).

    This file is part of cardinal_pythonlib.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        https://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

===============================================================================

**Support functions for Docker.**
"""

import os


def running_under_docker() -> bool:
    """
    Are we running inside a Docker container?

    As per
    https://stackoverflow.com/questions/43878953/how-does-one-detect-if-one-is-running-within-a-docker-container-within-python
    ... but without leaving a file open.
    """  # noqa
    # 1. Does /.dockerenv exist?
    if os.path.exists('/.dockerenv'):
        return True
    # 2. Is there a line containing "docker" in /proc/self/cgroup?
    path = '/proc/self/cgroup'
    if os.path.isfile(path):
        with open(path) as f:
            for line in f:
                if "docker" in line:
                    return True
    return False



#!/usr/bin/env python
# cardinal_pythonlib/version.py

"""
===============================================================================

    Original code copyright (C) 2009-2021 Rudolf Cardinal (rudolf@pobox.com).

    This file is part of cardinal_pythonlib.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        https://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

===============================================================================

**Functions to check the version of this library.**

"""

from semantic_version import Version

from cardinal_pythonlib.version_string import VERSION_STRING as VERSION

CARDINAL_PYTHONLIB_VERSION = Version(VERSION)


def _get_version_failure_msg(op: str, version_str: str) -> str:
    return f"Requested {op}{version_str} but cardinal_pythonlib=={VERSION}"


def assert_version_lt(version_str: str) -> None:
    """
    Asserts that the cardinal_pythonlib version is less than the value supplied
    (as a semantic version string, e.g. "1.0.2").
    """
    assert CARDINAL_PYTHONLIB_VERSION < Version(version_str), (
        _get_version_failure_msg("<", version_str)
    )


def assert_version_le(version_str: str) -> None:
    """
    Asserts that the cardinal_pythonlib version is less than or equal to the
    value supplied (as a semantic version string, e.g. "1.0.2").
    """
    assert CARDINAL_PYTHONLIB_VERSION <= Version(version_str), (
        _get_version_failure_msg("<=", version_str)
    )


def assert_version_eq(version_str: str) -> None:
    """
    Asserts that the cardinal_pythonlib version is equal to the value supplied
    (as a semantic version string, e.g. "1.0.2").
    """
    assert CARDINAL_PYTHONLIB_VERSION == Version(version_str), (
        _get_version_failure_msg("==", version_str)
    )


def assert_version_ge(version_str: str) -> None:
    """
    Asserts that the cardinal_pythonlib version is greater than or equal to the
    value supplied (as a semantic version string, e.g. "1.0.2").
    """
    assert CARDINAL_PYTHONLIB_VERSION >= Version(version_str), (
        _get_version_failure_msg(">=", version_str)
    )


def assert_version_gt(version_str: str) -> None:
    """
    Asserts that the cardinal_pythonlib version is greater than the value
    supplied (as a semantic version string, e.g. "1.0.2").
    """
    assert CARDINAL_PYTHONLIB_VERSION > Version(version_str), (
        _get_version_failure_msg(">", version_str)
    )



#!/usr/bin/env python
# cardinal_pythonlib/sqlalchemy/subproc.py

"""
===============================================================================

    Original code copyright (C) 2009-2021 Rudolf Cardinal (rudolf@pobox.com).

    This file is part of cardinal_pythonlib.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        https://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

===============================================================================
"""

import atexit
import logging
from multiprocessing.dummy import Pool  # thread pool
import os
from queue import Queue
from subprocess import (
    check_call,
    PIPE,
    Popen,
    TimeoutExpired,
)
import sys
from threading import Thread
from time import sleep
from typing import Any, BinaryIO, List, NoReturn, Optional, Tuple, Union

from cardinal_pythonlib.logs import get_brace_style_log_with_null_handler
from cardinal_pythonlib.cmdline import cmdline_quote

log = get_brace_style_log_with_null_handler(__name__)


# =============================================================================
# Constants and constant-like singletons
# =============================================================================

class SubprocSource(object):
    pass


SOURCE_STDOUT = SubprocSource()
SOURCE_STDERR = SubprocSource()


class SubprocCommand(object):
    pass


TERMINATE_SUBPROCESS = SubprocCommand()


# =============================================================================
# Processes that we're running
# =============================================================================

processes = []  # type: List[Popen]
proc_args_list = []  # type: List[List[str]]
# ... to report back which process failed, if any did


# =============================================================================
# Exiting
# =============================================================================

@atexit.register
def kill_child_processes() -> None:
    """
    Kills children of this process that were registered in the
    :data:`processes` variable.

    Use with ``@atexit.register``.
    """
    timeout_sec = 5
    for p in processes:
        try:
            p.wait(timeout_sec)
        except TimeoutExpired:
            # failed to close
            p.kill()  # you're dead


def fail() -> NoReturn:
    """
    Call when a child process has failed, and this will print an error
    message to ``stdout`` and execute ``sys.exit(1)`` (which will, in turn,
    call any ``atexit`` handler to kill children of this process).
    """
    print("\nPROCESS FAILED; EXITING ALL\n")
    sys.exit(1)  # will call the atexit handler and kill everything else


# =============================================================================
# Subprocess handling
# =============================================================================

def check_call_process(args: List[str]) -> None:
    """
    Logs the command arguments, then executes the command via
    :func:`subprocess.check_call`.
    """
    log.debug("{!r}", args)
    check_call(args)


def check_call_verbose(args: List[str],
                       log_level: Optional[int] = logging.INFO,
                       **kwargs) -> None:
    """
    Prints a copy/paste-compatible version of a command, then runs it.

    Args:
        args: command arguments
        log_level: log level

    Raises:
        :exc:`CalledProcessError` on external command failure
    """
    if log_level is not None:
        cmd_as_text = cmdline_quote(args)
        msg = f"[From directory {os.getcwd()}]: {cmd_as_text}"
        log.log(level=log_level, msg=msg)
    check_call(args, **kwargs)


def start_process(args: List[str],
                  stdin: Any = None,
                  stdout: Any = None,
                  stderr: Any = None) -> Popen:
    """
    Launch a child process and record it in our :data:`processes` variable.

    Args:
        args: program and its arguments, as a list
        stdin: typically None
        stdout: use None to perform no routing, which preserves console colour!
            Otherwise, specify somewhere to route stdout. See subprocess
            documentation. If either is PIPE, you'll need to deal with the
            output.
        stderr: As above. You can use stderr=STDOUT to route stderr to the same
            place as stdout.

    Returns:
        The process object (which is also stored in :data:`processes`).
    """
    log.debug("{!r}", args)
    global processes
    global proc_args_list
    proc = Popen(args, stdin=stdin, stdout=stdout, stderr=stderr)
    # proc = Popen(args, stdin=None, stdout=PIPE, stderr=STDOUT)
    # proc = Popen(args, stdin=None, stdout=PIPE, stderr=PIPE)
    # Can't preserve colour: https://stackoverflow.com/questions/13299550/preserve-colored-output-from-python-os-popen  # noqa
    processes.append(proc)
    proc_args_list.append(args)
    return proc


def wait_for_processes(die_on_failure: bool = True,
                       timeout_sec: float = 1) -> None:
    """
    Wait for child processes (catalogued in :data:`processes`) to finish.

    If ``die_on_failure`` is ``True``, then whenever a subprocess returns
    failure, all are killed.

    If ``timeout_sec`` is None, the function waits for its first process to
    complete, then waits for the second, etc. So a subprocess dying does not
    trigger a full quit instantly (or potentially for ages).

    If ``timeout_sec`` is something else, each process is tried for that time;
    if it quits within that time, well and good (successful quit -> continue
    waiting for the others; failure -> kill everything, if ``die_on_failure``);
    if it doesn't, we try the next. That is much more responsive.

    """
    global processes
    global proc_args_list
    n = len(processes)
    Pool(n).map(print_lines, processes)  # in case of PIPE
    something_running = True
    while something_running:
        something_running = False
        for i, p in enumerate(processes):
            try:
                retcode = p.wait(timeout=timeout_sec)
                if retcode == 0:
                    log.info("Process #{} (of {}) exited cleanly", i, n)
                if retcode != 0:
                    log.critical(
                        "Process #{} (of {}) exited with return code {} "
                        "(indicating failure); its args were: {!r}",
                        i, n, retcode, proc_args_list[i])
                    if die_on_failure:
                        log.critical("Exiting top-level process (will kill "
                                     "all other children)")
                        fail()  # exit this process, therefore kill its children  # noqa
            except TimeoutExpired:
                something_running = True
    processes.clear()
    proc_args_list.clear()


def print_lines(process: Popen) -> None:
    """
    Let a subprocess :func:`communicate`, then write both its ``stdout`` and
    its ``stderr`` to our ``stdout``.
    """
    out, err = process.communicate()
    if out:
        for line in out.decode("utf-8").splitlines():
            print(line)
    if err:
        for line in err.decode("utf-8").splitlines():
            print(line)


def run_multiple_processes(args_list: List[List[str]],
                           die_on_failure: bool = True) -> None:
    """
    Fire up multiple processes, and wait for them to finihs.

    Args:
        args_list: command arguments for each process
        die_on_failure: see :func:`wait_for_processes`
    """
    for procargs in args_list:
        start_process(procargs)
    # Wait for them all to finish
    wait_for_processes(die_on_failure=die_on_failure)


class AsynchronousFileReader(Thread):
    """
    Helper class to implement asynchronous reading of a file
    in a separate thread. Pushes read lines on a queue to
    be consumed in another thread.

    Modified from
    https://stefaanlippens.net/python-asynchronous-subprocess-pipe-reading/.
    """

    def __init__(self,
                 fd: BinaryIO,
                 queue: Queue,
                 encoding: str,
                 line_terminators: List[str] = None,
                 cmdargs: List[str] = None,
                 suppress_decoding_errors: bool = True) -> None:
        """
        Args:
            fd: file-like object to read from
            queue: queue to write to
            encoding: encoding to use when reading from the file
            line_terminators: valid line terminators
            cmdargs: for display purposes only: command that produced/is
                producing the file-like object
            suppress_decoding_errors: trap any ``UnicodeDecodeError``?
        """
        assert isinstance(queue, Queue)
        assert callable(fd.readline)
        super().__init__()
        self._fd = fd
        self._queue = queue
        self._encoding = encoding
        self._line_terminators = line_terminators or ["\n"]  # type: List[str]
        self._cmdargs = cmdargs or []  # type: List[str]
        self._suppress_decoding_errors = suppress_decoding_errors

    def run(self) -> None:
        """
        Read lines and put them on the queue.
        """
        fd = self._fd
        encoding = self._encoding
        line_terminators = self._line_terminators
        queue = self._queue
        buf = ""
        while True:
            try:
                c = fd.read(1).decode(encoding)
            except UnicodeDecodeError as e:
                log.warning("Decoding error from {!r}: {!r}", self._cmdargs, e)
                if self._suppress_decoding_errors:
                    continue
                else:
                    raise
            # log.critical("c={!r}, returncode={!r}", c, p.returncode)
            if not c:
                # Subprocess has finished
                return
            buf += c
            # log.critical("buf={!r}", buf)
            # noinspection PyTypeChecker
            for t in line_terminators:
                try:
                    t_idx = buf.index(t) + len(t)  # include terminator
                    fragment = buf[:t_idx]
                    buf = buf[t_idx:]
                    queue.put(fragment)
                except ValueError:
                    pass

    def eof(self) -> bool:
        """
        Check whether there is no more content to expect.
        """
        return not self.is_alive() and self._queue.empty()


def mimic_user_input(
        args: List[str],
        source_challenge_response: List[Tuple[SubprocSource,
                                              str,
                                              Union[str, SubprocCommand]]],
        line_terminators: List[str] = None,
        print_stdout: bool = False,
        print_stderr: bool = False,
        print_stdin: bool = False,
        stdin_encoding: str = None,
        stdout_encoding: str = None,
        suppress_decoding_errors: bool = True,
        sleep_time_s: float = 0.1) -> None:
    r"""
    Run an external command. Pretend to be a human by sending text to the
    subcommand (responses) when the external command sends us triggers
    (challenges).

    This is a bit nasty.

    Args:
        args: command-line arguments
        source_challenge_response: list of tuples of the format ``(challsrc,
            challenge, response)``; see below
        line_terminators: valid line terminators
        print_stdout:
        print_stderr:
        print_stdin:
        stdin_encoding:
        stdout_encoding:
        suppress_decoding_errors: trap any ``UnicodeDecodeError``?
        sleep_time_s:

    The ``(challsrc, challenge, response)`` tuples have this meaning:

    - ``challsrc``: where is the challenge coming from? Must be one of the
      objects :data:`SOURCE_STDOUT` or :data:`SOURCE_STDERR`;
    - ``challenge``: text of challenge
    - ``response``: text of response (send to the subcommand's ``stdin``).

    Example (modified from :class:`CorruptedZipReader`):

    .. code-block:: python

        from cardinal_pythonlib.subproc import *

        SOURCE_FILENAME = "corrupt.zip"
        TMP_DIR = "/tmp"
        OUTPUT_FILENAME = "rescued.zip"

        cmdargs = [
            "zip",  # Linux zip tool
            "-FF",  # or "--fixfix": "fix very broken things"
            SOURCE_FILENAME,  # input file
            "--temp-path", TMP_DIR,  # temporary storage path
            "--out", OUTPUT_FILENAME  # output file
        ]

        # We would like to be able to say "y" automatically to
        # "Is this a single-disk archive?  (y/n):"
        # The source code (api.c, zip.c, zipfile.c), from
        # ftp://ftp.info-zip.org/pub/infozip/src/ , suggests that "-q"
        # should do this (internally "-q" sets "noisy = 0") - but in
        # practice it doesn't work. This is a critical switch.
        # Therefore we will do something very ugly, and send raw text via
        # stdin.

        ZIP_PROMPTS_RESPONSES = [
            (SOURCE_STDOUT, "Is this a single-disk archive?  (y/n): ", "y\n"),
            (SOURCE_STDOUT, " or ENTER  (try reading this split again): ", "q\n"),
            (SOURCE_STDERR,
             "zip: malloc.c:2394: sysmalloc: Assertion `(old_top == initial_top (av) "
             "&& old_size == 0) || ((unsigned long) (old_size) >= MINSIZE && "
             "prev_inuse (old_top) && ((unsigned long) old_end & (pagesize - 1)) "
             "== 0)' failed.", TERMINATE_SUBPROCESS),
        ]
        ZIP_STDOUT_TERMINATORS = ["\n", "): "]

        mimic_user_input(cmdargs,
                         source_challenge_response=ZIP_PROMPTS_RESPONSES,
                         line_terminators=ZIP_STDOUT_TERMINATORS,
                         print_stdout=show_zip_output,
                         print_stdin=show_zip_output)

    """  # noqa
    line_terminators = line_terminators or ["\n"]  # type: List[str]
    stdin_encoding = stdin_encoding or sys.getdefaultencoding()
    stdout_encoding = stdout_encoding or sys.getdefaultencoding()

    # Launch the command
    p = Popen(args, stdin=PIPE, stdout=PIPE, stderr=PIPE, bufsize=0)

    # Launch the asynchronous readers of stdout and stderr
    stdout_queue = Queue()
    # noinspection PyTypeChecker
    stdout_reader = AsynchronousFileReader(
        fd=p.stdout,
        queue=stdout_queue,
        encoding=stdout_encoding,
        line_terminators=line_terminators,
        cmdargs=args,
        suppress_decoding_errors=suppress_decoding_errors
    )
    stdout_reader.start()
    stderr_queue = Queue()
    # noinspection PyTypeChecker
    stderr_reader = AsynchronousFileReader(
        fd=p.stderr,
        queue=stderr_queue,
        encoding=stdout_encoding,  # same as stdout
        line_terminators=line_terminators,
        cmdargs=args,
        suppress_decoding_errors=suppress_decoding_errors
    )
    stderr_reader.start()

    while not stdout_reader.eof() or not stderr_reader.eof():
        lines_with_source = []  # type: List[Tuple[SubprocSource, str]]
        while not stdout_queue.empty():
            lines_with_source.append((SOURCE_STDOUT, stdout_queue.get()))
        while not stderr_queue.empty():
            lines_with_source.append((SOURCE_STDERR, stderr_queue.get()))

        for src, line in lines_with_source:
            if src is SOURCE_STDOUT and print_stdout:
                print(line, end="")  # terminator already in line
            if src is SOURCE_STDERR and print_stderr:
                print(line, end="")  # terminator already in line
            for challsrc, challenge, response in source_challenge_response:
                # log.critical("challsrc={!r}", challsrc)
                # log.critical("challenge={!r}", challenge)
                # log.critical("line={!r}", line)
                # log.critical("response={!r}", response)
                if challsrc != src:
                    continue
                if challenge in line:
                    if response is TERMINATE_SUBPROCESS:
                        log.warning("Terminating subprocess {!r} because input "
                                    "{!r} received", args, challenge)
                        p.kill()
                        return
                    else:
                        p.stdin.write(response.encode(stdin_encoding))
                        p.stdin.flush()
                        if print_stdin:
                            print(response, end="")

        # Sleep a bit before asking the readers again.
        sleep(sleep_time_s)

    stdout_reader.join()
    stderr_reader.join()
    p.stdout.close()
    p.stderr.close()



#!/usr/bin/env python
# cardinal_pythonlib/dsp.py

"""
===============================================================================

    Original code copyright (C) 2009-2021 Rudolf Cardinal (rudolf@pobox.com).

    This file is part of cardinal_pythonlib.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        https://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

===============================================================================

**Digital signal processing functions.**

"""

from typing import List, Union

import numpy as np
from scipy.signal import firwin, iirnotch, lfilter

FLOATS_TYPE = Union[List[float], np.ndarray]


# =============================================================================
# Filters
# =============================================================================

def normalized_frequency(f: float, sampling_freq: float) -> float:
    """
    Returns a normalized frequency:

    Args:
        f: frequency :math:`f`
        sampling_freq: sampling frequency :math:`f_s`

    Returns:
        normalized frequency

            .. math::

                f_n = f / (f_s / 2)

    Principles:

    - if maximum frequency of interest is :math:`f`, then you should sample at
      the Nyquist rate of :math:`2f`;
    - if you sample at :math:`f_s`, then the maximum frequency is
      the Nyquist frequency :math:`0.5 f_s`
    - if you sample at :math:`2f`, then the normalized frequency is
      the range :math:`[0, 1]` for the frequency range :math:`[0, f]`.
    - e.g. see https://en.wikipedia.org/wiki/Nyquist_frequency,
      https://docs.scipy.org/doc/scipy/reference/generated/scipy.signal.iirnotch.html

    """  # noqa
    return f / (sampling_freq / 2.0)


def lowpass_filter(data: FLOATS_TYPE,
                   sampling_freq_hz: float,
                   cutoff_freq_hz: float,
                   numtaps: int) -> FLOATS_TYPE:
    """
    Apply a low-pass filter to the data.

    Args:
        data: time series of the data
        sampling_freq_hz: sampling frequency :math:`f_s`, in Hz
            (or other consistent units)
        cutoff_freq_hz: filter cutoff frequency in Hz
            (or other consistent units)
        numtaps: number of filter taps

    Returns:
        filtered data

    Note: number of filter taps = filter order + 1
    """
    # noinspection PyTypeChecker
    coeffs = firwin(
        numtaps=numtaps,
        cutoff=normalized_frequency(cutoff_freq_hz, sampling_freq_hz),
        pass_zero=True
    )  # coefficients of a finite impulse response (FIR) filter using window method  # noqa
    filtered_data = lfilter(b=coeffs, a=1.0, x=data)
    return filtered_data


def highpass_filter(data: FLOATS_TYPE,
                    sampling_freq_hz: float,
                    cutoff_freq_hz: float,
                    numtaps: int) -> FLOATS_TYPE:
    """
    Apply a high-pass filter to the data.

    Args:
        data: time series of the data
        sampling_freq_hz: sampling frequency :math:`f_s`, in Hz
            (or other consistent units)
        cutoff_freq_hz: filter cutoff frequency in Hz
            (or other consistent units)
        numtaps: number of filter taps

    Returns:
        filtered data

    Note: number of filter taps = filter order + 1
    """
    # noinspection PyTypeChecker
    coeffs = firwin(
        numtaps=numtaps,
        cutoff=normalized_frequency(cutoff_freq_hz, sampling_freq_hz),
        pass_zero=False
    )
    filtered_data = lfilter(b=coeffs, a=1.0, x=data)
    return filtered_data


def bandpass_filter(data: FLOATS_TYPE,
                    sampling_freq_hz: float,
                    lower_freq_hz: float,
                    upper_freq_hz: float,
                    numtaps: int) -> FLOATS_TYPE:
    """
    Apply a band-pass filter to the data.

    Args:
        data: time series of the data
        sampling_freq_hz: sampling frequency :math:`f_s`, in Hz
            (or other consistent units)
        lower_freq_hz: filter cutoff lower frequency in Hz
            (or other consistent units)
        upper_freq_hz: filter cutoff upper frequency in Hz
            (or other consistent units)
        numtaps: number of filter taps

    Returns:
        filtered data

    Note: number of filter taps = filter order + 1
    """
    f1 = normalized_frequency(lower_freq_hz, sampling_freq_hz)
    f2 = normalized_frequency(upper_freq_hz, sampling_freq_hz)
    # noinspection PyTypeChecker
    coeffs = firwin(
        numtaps=numtaps,
        cutoff=[f1, f2],
        pass_zero=False
    )
    filtered_data = lfilter(b=coeffs, a=1.0, x=data)
    return filtered_data


def notch_filter(data: FLOATS_TYPE,
                 sampling_freq_hz: float,
                 notch_freq_hz: float,
                 quality_factor: float) -> FLOATS_TYPE:
    """
    Design and use a notch (band reject) filter to filter the data.

    Args:
        data: time series of the data
        sampling_freq_hz: sampling frequency :math:`f_s`, in Hz
            (or other consistent units)
        notch_freq_hz: notch frequency, in Hz
            (or other consistent units)
        quality_factor: notch filter quality factor, :math:`Q`

    Returns:
        filtered data
    """
    b, a = iirnotch(
        w0=normalized_frequency(notch_freq_hz, sampling_freq_hz),
        Q=quality_factor
    )
    filtered_data = lfilter(b=b, a=a, x=data)
    return filtered_data



#!/usr/bin/env python
# cardinal_pythonlib/sql/literals.py

"""
===============================================================================

    Original code copyright (C) 2009-2021 Rudolf Cardinal (rudolf@pobox.com).

    This file is part of cardinal_pythonlib.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        https://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

===============================================================================

**Functions to check table/column names etc. for validity in SQL.**

This is a slight

"""

import re
from typing import Optional, Tuple


# =============================================================================
# SQL types and validation
# =============================================================================

# REGEX_INVALID_TABLE_FIELD_CHARS = re.compile("[^a-zA-Z0-9_ ]")
REGEX_INVALID_TABLE_FIELD_CHARS = re.compile("[^\x20-\x7E]")
# ... SQL Server is very liberal!


def is_valid_field_name(f: Optional[str]) -> bool:
    if not f:
        return False
    if bool(REGEX_INVALID_TABLE_FIELD_CHARS.search(f)):
        return False
    return True


def is_valid_table_name(t: Optional[str]) -> bool:
    return is_valid_field_name(t)


def ensure_valid_field_name(f: Optional[str]) -> None:
    if not is_valid_field_name(f):
        raise ValueError(f"Field name invalid: {f}")


def ensure_valid_table_name(t: Optional[str]) -> None:
    if not is_valid_table_name(t):
        raise ValueError(f"Table name invalid: {t}")


SQLTYPES_INTEGER = [
    "INT", "INTEGER",
    "TINYINT", "SMALLINT", "MEDIUMINT", "BIGINT",
]
SQLTYPES_FLOAT = [
    "DOUBLE", "FLOAT",
]
SQLTYPES_OTHER_NUMERIC = [
    "BIT", "BOOL", "BOOLEAN", "DEC", "DECIMAL",
]
SQLTYPES_TEXT = [
    "CHAR", "VARCHAR", "NVARCHAR",
    "TINYTEXT", "TEXT", "NTEXT", "MEDIUMTEXT", "LONGTEXT",
]
SQLTYPES_BINARY = [
    "BINARY", "BLOB", "IMAGE", "LONGBLOB", "VARBINARY",
]

SQLTYPES_WITH_DATE = [
    "DATE", "DATETIME", "TIME", "TIMESTAMP",
]
SQLTYPES_DATETIME_OTHER = [
    "TIME", "YEAR",
]
SQLTYPES_DATETIME_ALL = SQLTYPES_WITH_DATE + SQLTYPES_DATETIME_OTHER

SQLTYPES_ALL = (
    SQLTYPES_INTEGER +
    SQLTYPES_FLOAT +
    SQLTYPES_OTHER_NUMERIC +
    SQLTYPES_TEXT +
    SQLTYPES_BINARY +
    SQLTYPES_DATETIME_ALL
)
# Could be more comprehensive!

SQLTYPES_NOT_TEXT = (
    SQLTYPES_INTEGER +
    SQLTYPES_FLOAT +
    SQLTYPES_OTHER_NUMERIC +
    SQLTYPES_DATETIME_ALL
)
SQLTYPES_NUMERIC = (
    SQLTYPES_INTEGER +
    SQLTYPES_FLOAT +
    SQLTYPES_OTHER_NUMERIC
)


def split_long_sqltype(datatype_long: str) -> Tuple[str, Optional[int]]:
    datatype_short = datatype_long.split("(")[0].strip()
    find_open = datatype_long.find("(")
    find_close = datatype_long.find(")")
    if 0 <= find_open < find_close:
        try:
            length = int(datatype_long[find_open + 1:find_close])
        except (TypeError, IndexError, ValueError):  # e.g. for "VARCHAR(MAX)"
            length = None
    else:
        length = None
    return datatype_short, length


def is_sqltype_valid(datatype_long: str) -> bool:
    (datatype_short, length) = split_long_sqltype(datatype_long)
    return datatype_short in SQLTYPES_ALL


def is_sqltype_date(datatype_long: str) -> bool:
    (datatype_short, length) = split_long_sqltype(datatype_long)
    return datatype_short in SQLTYPES_WITH_DATE


def is_sqltype_text(datatype_long: str) -> bool:
    (datatype_short, length) = split_long_sqltype(datatype_long)
    return datatype_short in SQLTYPES_TEXT


def is_sqltype_text_of_length_at_least(datatype_long: str,
                                       min_length: int) -> bool:
    (datatype_short, length) = split_long_sqltype(datatype_long)
    if datatype_short not in SQLTYPES_TEXT:
        return False
    if length is None:  # text, with no length, e.g. VARCHAR(MAX)
        return True
    return length >= min_length


def is_sqltype_text_over_one_char(datatype_long: str) -> bool:
    return is_sqltype_text_of_length_at_least(datatype_long, 2)


def is_sqltype_binary(datatype_long: str) -> bool:
    (datatype_short, length) = split_long_sqltype(datatype_long)
    return datatype_short in SQLTYPES_BINARY


def is_sqltype_numeric(datatype_long: str) -> bool:
    (datatype_short, length) = split_long_sqltype(datatype_long)
    return datatype_short in SQLTYPES_NUMERIC


def is_sqltype_integer(datatype_long: str) -> bool:
    (datatype_short, length) = split_long_sqltype(datatype_long)
    return datatype_short in SQLTYPES_INTEGER


def does_sqltype_require_index_len(datatype_long: str) -> bool:
    (datatype_short, length) = split_long_sqltype(datatype_long)
    return datatype_short in ["TEXT", "BLOB"]


def does_sqltype_merit_fulltext_index(datatype_long: str,
                                      min_length: int = 1000) -> bool:
    return is_sqltype_text_of_length_at_least(datatype_long, min_length)



#!/usr/bin/env python
# cardinal_pythonlib/psychiatry/timeline.py

"""
===============================================================================

    Original code copyright (C) 2009-2021 Rudolf Cardinal (rudolf@pobox.com).

    This file is part of cardinal_pythonlib.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        https://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

===============================================================================

Timeline calculations.
Primarily for a lithium/renal function project, Apr 2019.
Code is in DRAFT.

Usage from R:

.. code-block:: r

    # -------------------------------------------------------------------------
    # Load libraries
    # -------------------------------------------------------------------------

    RUN_ONCE_ONLY <- '
        library(devtools)
        devtools::install_github("rstudio/reticulate")  # get latest version
    '
    library(data.table)
    library(reticulate)

    # -------------------------------------------------------------------------
    # Set up reticulate
    # -------------------------------------------------------------------------

    VENV <- "~/dev/venvs/cardinal_pythonlib"  # or your preferred virtualenv
    PYTHON_EXECUTABLE <- ifelse(
        .Platform$OS.type == "windows",
        file.path(VENV, "Scripts", "python.exe"),  # Windows
        file.path(VENV, "bin", "python")  # Linux
    )
    reticulate::use_python(PYTHON_EXECUTABLE, required=TRUE)

    # -------------------------------------------------------------------------
    # Import Python modules
    # -------------------------------------------------------------------------

    cpl_version <- reticulate::import("cardinal_pythonlib.version")
    cpl_version$assert_version_eq("1.0.50")
    cpl_logs <- reticulate::import("cardinal_pythonlib.logs")
    cpl_logs$main_only_quicksetup_rootlogger()

    cpl_timeline <- reticulate::import("cardinal_pythonlib.psychiatry.timeline")

    # -------------------------------------------------------------------------
    # Do something
    # -------------------------------------------------------------------------

    testdata_drug_events <- data.table(
        patient_id=c(
            rep("Alice", 3),
            rep("Bob", 3)
        ),
        drug_event_datetime=as.Date(c(
            # Alice
            "2018-01-05",
            "2018-01-20",
            "2018-04-01",
            # Bob
            "2018-06-05",
            "2018-08-20",
            "2018-10-01"
        ))
    )
    testdata_query_times <- data.table(
        patient_id=c(
            rep("Alice", 3),
            rep("Bob", 3)
        ),
        start=as.Date(c(
            # Alice
            rep("2017-01-01", 3),
            # Bob
            rep("2015-01-01", 3)
        )),
        when=as.Date(c(
            # Alice
            "2018-01-01",
            "2018-01-10",
            "2018-02-01",
            # Bob
            "2018-01-01",
            "2018-09-10",
            "2019-02-01"
        ))
    )
    testresult <- data.table(cpl_timeline$cumulative_time_on_drug(
        drug_events_df=testdata_drug_events,
        event_lasts_for_quantity=3,
        event_lasts_for_units="days",
        query_times_df=testdata_query_times,
        patient_colname="patient_id",
        event_datetime_colname="drug_event_datetime",
        start_colname="start",
        when_colname="when",
        debug=TRUE
    ))
    print(testresult)

The result should be:

.. code-block:: none

    > print(testdata_drug_events)

       patient_id drug_event_datetime
    1:      Alice          2018-01-05
    2:      Alice          2018-01-20
    3:      Alice          2018-04-01
    4:        Bob          2018-06-05
    5:        Bob          2018-08-20
    6:        Bob          2018-10-01

    > print(testdata_query_times)

       patient_id      start       when
    1:      Alice 2017-01-01 2018-01-01
    2:      Alice 2017-01-01 2018-01-10
    3:      Alice 2017-01-01 2018-02-01
    4:        Bob 2015-01-01 2018-01-01
    5:        Bob 2015-01-01 2018-09-10
    6:        Bob 2015-01-01 2019-02-01

    > print(testresult)

       patient_id      start          t before_days during_days after_days
    1:      Alice 2017-01-01 2018-01-01         365           0          0
    2:      Alice 2017-01-01 2018-01-10         369           3          2
    3:      Alice 2017-01-01 2018-02-01         369           6         21
    4:        Bob 2015-01-01 2018-01-01        1096           0          0
    5:        Bob 2015-01-01 2018-09-10        1251           6         91
    6:        Bob 2015-01-01 2019-02-01        1251           9        232

However, there is a ``reticulate`` bug that can cause problems, by corrupting
dates passed from R to Python:

.. code-block:: r

    # PROBLEM on 2018-04-05, with reticulate 1.11.1:
    # - the R data.table is fine
    # - all the dates become the same date when it's seen by Python (the value
    #   of the first row in each date column)
    # - when used without R, the Python code is fine
    # - therefore, a problem with reticulate converting data for Python
    # - same with data.frame() as with data.table()
    # - same with as.Date.POSIXct() and as.Date.POSIXlt() as with as.Date()

    # Further test:

    cpl_rfunc <- reticulate::import("cardinal_pythonlib.psychiatry.rfunc")
    cat(cpl_rfunc$get_python_repr(testdata_drug_events))
    cat(cpl_rfunc$get_python_repr_of_type(testdata_drug_events))
    print(testdata_drug_events)
    print(reticulate::r_to_py(testdata_drug_events))

    # Minimum reproducible example:

    library(reticulate)
    testdata_drug_events <- data.frame(
        patient_id=c(
            rep("Alice", 3),
            rep("Bob", 3)
        ),
        drug_event_datetime=as.Date(c(
            # Alice
            "2018-01-05",
            "2018-01-20",
            "2018-04-01",
            # Bob
            "2018-06-05",
            "2018-08-20",
            "2018-10-01"
        ))
    )
    print(testdata_drug_events)
    print(reticulate::r_to_py(testdata_drug_events))

    # The R data is:
    #
    #       patient_id drug_event_datetime
    #     1      Alice          2018-01-05
    #     2      Alice          2018-01-20
    #     3      Alice          2018-04-01
    #     4        Bob          2018-06-05
    #     5        Bob          2018-08-20
    #     6        Bob          2018-10-01
    #
    # Output from reticulate::r_to_py() in the buggy version is:
    #
    #       patient_id drug_event_datetime
    #     0      Alice          2018-01-05
    #     1      Alice          2018-01-05
    #     2      Alice          2018-01-05
    #     3        Bob          2018-01-05
    #     4        Bob          2018-01-05
    #     5        Bob          2018-01-05
    #
    # Known bug: https://github.com/rstudio/reticulate/issues/454
    #
    # Use remove.packages() then reinstall from github as above, giving
    # reticulate_1.11.1-9000 [see sessionInfo()]...
    # ... yes, that fixes it.

"""

from collections import defaultdict
import datetime
import logging
import sys
from typing import Any, Dict, List
import unittest

from numpy import array
from pandas import DataFrame

from cardinal_pythonlib.interval import Interval, IntervalList
from cardinal_pythonlib.logs import (
    BraceStyleAdapter,
    main_only_quicksetup_rootlogger,
)

log = BraceStyleAdapter(logging.getLogger(__name__))

DEFAULT_PATIENT_COLNAME = "patient_id"
DEFAULT_DRUG_EVENT_DATETIME_COLNAME = "drug_event_datetime"
DEFAULT_START_DATETIME_COLNAME = "start"
DEFAULT_QUERY_DATETIME_COLNAME = "when"


def drug_timelines(
        drug_events_df: DataFrame,
        event_lasts_for: datetime.timedelta,
        patient_colname: str = DEFAULT_PATIENT_COLNAME,
        event_datetime_colname: str = DEFAULT_DRUG_EVENT_DATETIME_COLNAME) \
        -> Dict[Any, IntervalList]:
    """
    Takes a set of drug event start times (one or more per patient), plus a
    fixed time that each event is presumed to last for, and returns an
    :class:`IntervalList` for each patient representing the set of events
    (which may overlap, in which case they will be amalgamated).

    Args:
        drug_events_df:
            pandas :class:`DataFrame` containing the event data
        event_lasts_for:
            when an event occurs, how long is it assumed to last for? For
            example, if a prescription of lithium occurs on 2001-01-01, how
            long is the patient presumed to be taking lithium as a consequence
            (e.g. 1 day? 28 days? 6 months?)
        patient_colname:
            name of the column in ``drug_events_df`` containing the patient ID
        event_datetime_colname:
            name of the column in ``drug_events_df`` containing the date/time
            of each event

    Returns:
        dict: mapping patient ID to a :class:`IntervalList` object indicating
        the amalgamated intervals from the events

    """
    sourcecolnum_pt = drug_events_df.columns.get_loc(patient_colname)
    sourcecolnum_when = drug_events_df.columns.get_loc(event_datetime_colname)
    timelines = defaultdict(IntervalList)
    nrows = len(drug_events_df)
    for rowidx in range(nrows):
        patient_id = drug_events_df.iat[rowidx, sourcecolnum_pt]
        event_when = drug_events_df.iat[rowidx, sourcecolnum_when]
        interval = Interval(event_when, event_when + event_lasts_for)
        ivlist = timelines[patient_id]  # will create if unknown
        ivlist.add(interval)
    return timelines


DTYPE_STRING = "<U255"
# ... see treatment_resistant_depression.py
DTYPE_DATETIME = "datetime64[s]"
# ... https://docs.scipy.org/doc/numpy/reference/arrays.datetime.html
DTYPE_FLOAT = "Float64"
# ... https://docs.scipy.org/doc/numpy/reference/arrays.dtypes.html
DTYPE_TIMEDELTA = "timedelta64"

RCN_PATIENT_ID = "patient_id"  # RCN: "result column name"
RCN_START = "start"
RCN_TIME = "t"
RCN_BEFORE_TIMEDELTA = "before_timedelta"
RCN_DURING_TIMEDELTA = "during_timedelta"
RCN_AFTER_TIMEDELTA = "after_timedelta"
RCN_BEFORE_DAYS = "before_days"
RCN_DURING_DAYS = "during_days"
RCN_AFTER_DAYS = "after_days"


def cumulative_time_on_drug(
        drug_events_df: DataFrame,
        query_times_df: DataFrame,
        event_lasts_for_timedelta: datetime.timedelta = None,
        event_lasts_for_quantity: float = None,
        event_lasts_for_units: str = None,
        patient_colname: str = DEFAULT_PATIENT_COLNAME,
        event_datetime_colname: str = DEFAULT_DRUG_EVENT_DATETIME_COLNAME,
        start_colname: str = DEFAULT_START_DATETIME_COLNAME,
        when_colname: str = DEFAULT_QUERY_DATETIME_COLNAME,
        include_timedelta_in_output: bool = False,
        debug: bool = False) \
        -> DataFrame:
    """

    Args:
        drug_events_df:
            pandas :class:`DataFrame` containing the event data, with columns
            named according to ``patient_colname``, ``event_datetime_colname``
        event_lasts_for_timedelta:
            when an event occurs, how long is it assumed to last for? For
            example, if a prescription of lithium occurs on 2001-01-01, how
            long is the patient presumed to be taking lithium as a consequence
            (e.g. 1 day? 28 days? 6 months?)
        event_lasts_for_quantity:
            as an alternative to ``event_lasts_for_timedelta``, particularly if
            you are calling from R to Python via ``reticulate`` (which doesn't
            convert R ``as.difftime()`` to Python ``datetime.timedelta``), you
            can specify ``event_lasts_for_quantity``, a number and
            ``event_lasts_for_units`` (q.v.).
        event_lasts_for_units:
            specify the units for ``event_lasts_for_quantity`` (q.v.), if used;
            e.g. ``"days"``. The string value must be the name of an argument
            to the Python ``datetime.timedelta`` constructor.
        query_times_df:
            times to query for, with columns named according to
            ``patient_colname``, ``start_colname``, and ``when_colname``
        patient_colname:
            name of the column in ``drug_events_df`` and ``query_time_df``
            containing the patient ID
        event_datetime_colname:
            name of the column in ``drug_events_df`` containing the date/time
            of each event
        start_colname:
            name of the column in ``query_time_df`` containing the date/time
            representing the overall start time for the relevant patient (from
            which cumulative times are calculated)
        when_colname:
            name of the column in ``query_time_df`` containing date/time
            values at which to query
        include_timedelta_in_output:
            include ``datetime.timedelta`` values in the output? The default is
            ``False`` as this isn't supported by R/``reticulate``.
        debug:
            print debugging information to the log?

    Returns:
        :class:`DataFrame` with the requested data

    """
    if event_lasts_for_timedelta is None:
        assert event_lasts_for_quantity and event_lasts_for_units
        timedelta_dict = {event_lasts_for_units: event_lasts_for_quantity}
        event_lasts_for_timedelta = datetime.timedelta(**timedelta_dict)
    if debug:
        log.critical("drug_events_df:\n{!r}", drug_events_df)
        log.critical("event_lasts_for:\n{!r}", event_lasts_for_timedelta)
        log.critical("query_times_df:\n{!r}", query_times_df)
    timelines = drug_timelines(
        drug_events_df=drug_events_df,
        event_lasts_for=event_lasts_for_timedelta,
        patient_colname=patient_colname,
        event_datetime_colname=event_datetime_colname,
    )
    query_nrow = len(query_times_df)
    ct_coldefs = [  # column definitions:
        (RCN_PATIENT_ID, DTYPE_STRING),
        (RCN_START, DTYPE_DATETIME),
        (RCN_TIME, DTYPE_DATETIME),
        (RCN_BEFORE_DAYS, DTYPE_FLOAT),
        (RCN_DURING_DAYS, DTYPE_FLOAT),
        (RCN_AFTER_DAYS, DTYPE_FLOAT),
    ]
    if include_timedelta_in_output:
        ct_coldefs.extend([
            (RCN_BEFORE_TIMEDELTA, DTYPE_TIMEDELTA),
            (RCN_DURING_TIMEDELTA, DTYPE_TIMEDELTA),
            (RCN_AFTER_TIMEDELTA, DTYPE_TIMEDELTA),
        ])
    ct_arr = array([None] * query_nrow, dtype=ct_coldefs)
    # log.debug("ct_arr:\n{!r}", ct_arr)
    cumulative_times = DataFrame(ct_arr, index=list(range(query_nrow)))
    # log.debug("cumulative_times:\n{!r}", cumulative_times)
    # So we can use the fast "iat" function.
    sourcecolnum_pt = query_times_df.columns.get_loc(patient_colname)
    sourcecolnum_start = query_times_df.columns.get_loc(start_colname)
    sourcecolnum_when = query_times_df.columns.get_loc(when_colname)
    dest_colnum_pt = cumulative_times.columns.get_loc(RCN_PATIENT_ID)
    dest_colnum_start = cumulative_times.columns.get_loc(RCN_START)
    dest_colnum_t = cumulative_times.columns.get_loc(RCN_TIME)
    dest_colnum_before_days = cumulative_times.columns.get_loc(RCN_BEFORE_DAYS)
    dest_colnum_during_days = cumulative_times.columns.get_loc(RCN_DURING_DAYS)
    dest_colnum_after_days = cumulative_times.columns.get_loc(RCN_AFTER_DAYS)
    if include_timedelta_in_output:
        dest_colnum_before_dt = cumulative_times.columns.get_loc(RCN_BEFORE_TIMEDELTA)  # noqa
        dest_colnum_during_dt = cumulative_times.columns.get_loc(RCN_DURING_TIMEDELTA)  # noqa
        dest_colnum_after_dt = cumulative_times.columns.get_loc(RCN_AFTER_TIMEDELTA)  # noqa
    else:
        # for type checker
        dest_colnum_before_dt = 0
        dest_colnum_during_dt = 0
        dest_colnum_after_dt = 0
    for rowidx in range(query_nrow):
        patient_id = query_times_df.iat[rowidx, sourcecolnum_pt]
        start = query_times_df.iat[rowidx, sourcecolnum_start]
        when = query_times_df.iat[rowidx, sourcecolnum_when]
        ivlist = timelines[patient_id]
        # log.critical("ivlist: {!r}", ivlist)
        before, during, after = ivlist.cumulative_before_during_after(start,
                                                                      when)
        # log.critical(
        #     "{!r}.cumulative_before_during_after(start={!r}, when={!r}) "
        #     "-> {!r}, {!r}, {!r}",
        #     ivlist, start, when,
        #     before, during, after
        # )
        cumulative_times.iat[rowidx, dest_colnum_pt] = patient_id
        cumulative_times.iat[rowidx, dest_colnum_start] = start
        cumulative_times.iat[rowidx, dest_colnum_t] = when
        cumulative_times.iat[rowidx, dest_colnum_before_days] = before.days
        cumulative_times.iat[rowidx, dest_colnum_during_days] = during.days
        cumulative_times.iat[rowidx, dest_colnum_after_days] = after.days
        if include_timedelta_in_output:
            cumulative_times.iat[rowidx, dest_colnum_before_dt] = before
            cumulative_times.iat[rowidx, dest_colnum_during_dt] = during
            cumulative_times.iat[rowidx, dest_colnum_after_dt] = after
    return cumulative_times


# =============================================================================
# Unit testing
# =============================================================================

class TestTimeline(unittest.TestCase):
    """
    Unit tests.
    """
    DATEFORMAT = "%Y-%m-%d"
    DATETIMEFORMAT = "%Y-%m-%d %H:%M"

    DRUG_EVENT_TIME = " 00:00"  # " 09:00"
    QUERY_EVENT_TIME = " 00:00"  # " 12:00"

    @classmethod
    def dateseq(cls, first: str, last: str,
                time_suffix: str = "") -> List[datetime.datetime]:
        fmt = cls.DATETIMEFORMAT if time_suffix else cls.DATEFORMAT
        if time_suffix:
            first += time_suffix
            last += time_suffix
        dfirst = datetime.datetime.strptime(first, fmt)
        dlast = datetime.datetime.strptime(last, fmt)
        assert dfirst <= dlast
        dates = []  # type: List[datetime.datetime]
        d = dfirst
        while d <= dlast:
            dates.append(d)
            d += datetime.timedelta(days=1)
        return dates

    def test_timeline(self) -> None:
        event_lasts_for = datetime.timedelta(weeks=4)
        # event_lasts_for = datetime.timedelta(days=3)
        log.debug("event_lasts_for: {!r}", event_lasts_for)

        alice = "Alice"
        drug_events_arr = array(
            [
                # Alice
                (alice, "2018-01-05" + self.DRUG_EVENT_TIME),
                (alice, "2018-01-20" + self.DRUG_EVENT_TIME),
                (alice, "2018-04-01" + self.DRUG_EVENT_TIME),
            ],
            dtype=[
                (DEFAULT_PATIENT_COLNAME, DTYPE_STRING),
                (DEFAULT_DRUG_EVENT_DATETIME_COLNAME, DTYPE_DATETIME),
            ]
        )
        drug_events_df = DataFrame.from_records(drug_events_arr)
        log.debug("drug_events_df:\n{!r}", drug_events_df)

        start = datetime.datetime.strptime("2017-01-01" + self.DRUG_EVENT_TIME,
                                           self.DATETIMEFORMAT)
        log.debug("start: {!r}", start)

        qdata_rows = []
        for dt in self.dateseq("2018-01-01", "2018-05-30",
                               time_suffix=self.QUERY_EVENT_TIME):
            qdata_rows.append((alice, start, dt))
        query_times_arr = array(
            qdata_rows,
            dtype=[
                (DEFAULT_PATIENT_COLNAME, DTYPE_STRING),
                (DEFAULT_START_DATETIME_COLNAME, DTYPE_DATETIME),
                (DEFAULT_QUERY_DATETIME_COLNAME, DTYPE_DATETIME),
            ]
        )
        query_times_df = DataFrame.from_records(query_times_arr)
        log.debug("query_times_df:\n{!r}", query_times_df)

        timelines = drug_timelines(
            drug_events_df=drug_events_df,
            event_lasts_for=event_lasts_for,
            patient_colname=DEFAULT_PATIENT_COLNAME,
            event_datetime_colname=DEFAULT_DRUG_EVENT_DATETIME_COLNAME
        )
        log.debug("timelines: {!r}", timelines)

        cumulative = cumulative_time_on_drug(
            drug_events_df=drug_events_df,
            event_lasts_for_timedelta=event_lasts_for,
            query_times_df=query_times_df,
            patient_colname=DEFAULT_PATIENT_COLNAME,
            event_datetime_colname=DEFAULT_DRUG_EVENT_DATETIME_COLNAME,
            start_colname=DEFAULT_START_DATETIME_COLNAME,
            when_colname=DEFAULT_QUERY_DATETIME_COLNAME
        )
        log.debug("cumulative:\n{}", cumulative)


# =============================================================================
# main
# =============================================================================

if __name__ == "__main__":
    main_only_quicksetup_rootlogger(level=logging.DEBUG)
    log.info("Running unit tests")
    unittest.main(argv=[sys.argv[0]])
    sys.exit(0)



#!/usr/bin/env python3
# cardinal_pythonlib/openxml/find_recovered_openxml.py

"""
===============================================================================

    Original code copyright (C) 2009-2021 Rudolf Cardinal (rudolf@pobox.com).

    This file is part of cardinal_pythonlib.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        https://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

===============================================================================

**Tool to recognize and rescue Microsoft Office OpenXML files, even if they
have garbage appended to them. See the command-line help for details.**

Version history:

- Written 28 Sep 2017.

Notes:

- use the ``vbindiff`` tool to show *how* two binary files differ.

Output from ``zip -FF bad.zip --out good.zip``

.. code-block:: none

    Fix archive (-FF) - salvage what can
        zip warning: Missing end (EOCDR) signature - either this archive
                         is not readable or the end is damaged
    Is this a single-disk archive?  (y/n):

... and note there are some tabs in that, too.

More ``zip -FF`` output:

.. code-block:: none

    Fix archive (-FF) - salvage what can
     Found end record (EOCDR) - says expect 50828 splits
      Found archive comment
    Scanning for entries...


    Could not find:
      /home/rudolf/tmp/ziptest/00008470.z01

    Hit c      (change path to where this split file is)
        s      (skip this split)
        q      (abort archive - quit)
        e      (end this archive - no more splits)
        z      (look for .zip split - the last split)
     or ENTER  (try reading this split again):


More ``zip -FF`` output:

.. code-block:: none

    zip: malloc.c:2394: sysmalloc: ...

... this heralds a crash in ``zip``. We need to kill it; otherwise it just sits
there doing nothing and not asking for any input. Presumably this means the
file is badly corrupted (or not a zip at all).

"""

from argparse import ArgumentParser, RawDescriptionHelpFormatter
import fnmatch
import logging
import multiprocessing
import os
import re
import shutil
import struct
import tempfile
from time import sleep
import traceback
from typing import List
from zipfile import BadZipFile, ZipFile

from cardinal_pythonlib.logs import (
    BraceStyleAdapter,
    main_only_quicksetup_rootlogger,
)
from cardinal_pythonlib.fileops import exists_locked, gen_filenames
from cardinal_pythonlib.subproc import (
    mimic_user_input,
    SOURCE_STDERR,
    SOURCE_STDOUT,
    TERMINATE_SUBPROCESS,
)

log = BraceStyleAdapter(logging.getLogger(__name__))

DOCX_CONTENTS_REGEX_STR = "word/.*xml"
PPTX_CONTENTS_REGEX_STR = "ppt/.*xml"
XLSX_CONTENTS_REGEX_STR = "xl/.*xml"

DOCX_CONTENTS_REGEX = re.compile(DOCX_CONTENTS_REGEX_STR)
PPTX_CONTENTS_REGEX = re.compile(PPTX_CONTENTS_REGEX_STR)
XLSX_CONTENTS_REGEX = re.compile(XLSX_CONTENTS_REGEX_STR)

DOCX = "docx"
PPTX = "pptx"
XLSX = "xlsx"
FILETYPES = [DOCX, PPTX, XLSX]

ZIP_PROMPTS_RESPONSES = [
    (SOURCE_STDOUT, "Is this a single-disk archive?  (y/n): ", "y\n"),
    (SOURCE_STDOUT, " or ENTER  (try reading this split again): ", "q\n"),
    (SOURCE_STDERR,
     "zip: malloc.c:2394: sysmalloc: Assertion `(old_top == initial_top (av) "
     "&& old_size == 0) || ((unsigned long) (old_size) >= MINSIZE && "
     "prev_inuse (old_top) && ((unsigned long) old_end & (pagesize - 1)) "
     "== 0)' failed.", TERMINATE_SUBPROCESS),
]
ZIP_STDOUT_TERMINATORS = ["\n", "): "]


class CorruptedZipReader(object):
    """
    Class to open a zip file, even one that is corrupted, and detect the
    files within.
    """
    def __init__(self, filename: str, show_zip_output: bool = False) -> None:
        """
        Args:
            filename: filename of the ``.zip`` file (or corrupted ``.zip``
                file) to open
            show_zip_output: show the output of the external ``zip`` tool?
        """
        self.src_filename = filename
        self.rescue_filename = ""
        self.tmp_dir = ""
        self.contents_filenames = []  # type: List[str]

        try:
            # A happy zip file will be readable like this:
            with ZipFile(self.src_filename, 'r') as zip_ref:
                self.contents_filenames = zip_ref.namelist()
        except (BadZipFile, OSError) as e:
            # Here we have an unhappy zip file.
            log.debug("File {!r} raised error: {!r}", filename, e)
            self._fix_zip(show_zip_output=show_zip_output)
            try:
                with ZipFile(self.rescue_filename, 'r') as zip_ref:
                    self.contents_filenames = zip_ref.namelist()
            except (BadZipFile, OSError, struct.error) as e:
                log.debug("... exception raised even after fix attempt: {!r}",
                          e)
            if self.contents_filenames:
                log.debug("... recovered!")
            else:
                log.debug("... attempt at recovery failed")

    def _fix_zip(self, show_zip_output: bool = False) -> None:
        # We are trying to deal with ZIP (specifically, PPTX) files that
        # have been retrieved by Scalpel so have large extra bits of junk
        # on the end.
        # Make a file in a temporary directory
        self.tmp_dir = tempfile.mkdtemp()
        self.rescue_filename = os.path.join(
            self.tmp_dir, os.path.basename(self.src_filename))
        cmdargs = [
            "zip",  # Linux zip tool
            "-FF",  # or "--fixfix": "fix very broken things"
            self.src_filename,  # input file
            "--temp-path", self.tmp_dir,  # temporary storage path
            "--out", self.rescue_filename  # output file
        ]
        # We would like to be able to say "y" automatically to
        # "Is this a single-disk archive?  (y/n):"
        # The source code (api.c, zip.c, zipfile.c), from
        # ftp://ftp.info-zip.org/pub/infozip/src/ , suggests that "-q"
        # should do this (internally "-q" sets "noisy = 0") - but in
        # practice it doesn't work. This is a critical switch.
        # Therefore we will do something very ugly, and send raw text via
        # stdin.
        log.debug("Running {!r}", cmdargs)
        mimic_user_input(cmdargs,
                         source_challenge_response=ZIP_PROMPTS_RESPONSES,
                         line_terminators=ZIP_STDOUT_TERMINATORS,
                         print_stdout=show_zip_output,
                         print_stdin=show_zip_output)
        # ... will raise if the 'zip' tool isn't available

    def move_to(self, destination_filename: str,
                alter_if_clash: bool = True) -> None:
        """
        Move the file to which this class refers to a new location.
        The function will not overwrite existing files (but offers the option
        to rename files slightly to avoid a clash).

        Args:
            destination_filename: filename to move to
            alter_if_clash: if ``True`` (the default), appends numbers to
                the filename if the destination already exists, so that the
                move can proceed.
        """
        if not self.src_filename:
            return
        if alter_if_clash:
            counter = 0
            while os.path.exists(destination_filename):
                root, ext = os.path.splitext(destination_filename)
                destination_filename = f"{root}_{counter}{ext}"
                counter += 1
            # ... for example, "/a/b/c.txt" becomes "/a/b/c_0.txt", then
            # "/a/b/c_1.txt", and so on.
        else:
            if os.path.exists(destination_filename):
                src = self.rescue_filename or self.src_filename
                log.warning("Destination exists; won't move {!r} to {!r}",
                            src, destination_filename)
                return
        if self.rescue_filename:
            shutil.move(self.rescue_filename, destination_filename)
            os.remove(self.src_filename)
            log.info("Moved recovered file {!r} to {!r} and deleted corrupted "
                     "original {!r}",
                     self.rescue_filename,
                     destination_filename,
                     self.src_filename)
            self.rescue_filename = ""
        else:
            shutil.move(self.src_filename, destination_filename)
            log.info("Moved {!r} to {!r}", self.src_filename,
                     destination_filename)
        self.src_filename = ""

    def __del__(self) -> None:
        if self.tmp_dir:
            shutil.rmtree(self.tmp_dir)


class CorruptedOpenXmlReader(CorruptedZipReader):
    """
    Class to read a potentially corrupted OpenXML file.
    As it is created, it sets its ``file_type`` member to the detected OpenXML
    file type, if it can.
    """
    def __init__(self, filename: str, show_zip_output: bool = False) -> None:
        super().__init__(filename=filename,
                         show_zip_output=show_zip_output)
        self.file_type = ""
        self._recognize()

    def _recognize(self) -> None:
        for fname in self.contents_filenames:
            if DOCX_CONTENTS_REGEX.match(fname):
                log.debug("Zip file {!r} has Word DOCX contents {!r}",
                          self.src_filename, fname)
                self.file_type = DOCX
                return
            if PPTX_CONTENTS_REGEX.match(fname):
                log.debug("Zip file {!r} has Powerpoint PPTX contents {!r}",
                          self.src_filename, fname)
                self.file_type = PPTX
                return
            if XLSX_CONTENTS_REGEX.match(fname):
                log.debug("Zip file {!r} has Excel XLSX contents {!r}",
                          self.src_filename, fname)
                self.file_type = XLSX
                return

    def suggested_extension(self) -> str:
        if not self.file_type:
            return ""
        return "." + self.file_type

    @property
    def recognized(self) -> bool:
        return bool(self.file_type)

    @property
    def description(self) -> str:
        return self.file_type.upper()


def process_file(filename: str,
                 filetypes: List[str],
                 move_to: str,
                 delete_if_not_specified_file_type: bool,
                 show_zip_output: bool) -> None:
    """
    Deals with an OpenXML, including if it is potentially corrupted.

    Args:
        filename: filename to process
        filetypes: list of filetypes that we care about, e.g.
            ``['docx', 'pptx', 'xlsx']``.
        move_to: move matching files to this directory
        delete_if_not_specified_file_type: if ``True``, and the file is **not**
            a type specified in ``filetypes``, then delete the file.
        show_zip_output: show the output from the external ``zip`` tool?
    """
    # log.critical("process_file: start")
    try:
        reader = CorruptedOpenXmlReader(filename,
                                        show_zip_output=show_zip_output)
        if reader.file_type in filetypes:
            log.info("Found {}: {}", reader.description, filename)
            if move_to:
                dest_file = os.path.join(move_to, os.path.basename(filename))
                _, ext = os.path.splitext(dest_file)
                if ext != reader.suggested_extension():
                    dest_file += reader.suggested_extension()
                reader.move_to(destination_filename=dest_file)
        else:
            log.info("Unrecognized or unwanted contents: " + filename)
            if delete_if_not_specified_file_type:
                log.info("Deleting: " + filename)
                os.remove(filename)
    except Exception as e:
        # Must explicitly catch and report errors, since otherwise they vanish
        # into the ether.
        log.critical("Uncaught error in subprocess: {!r}\n{}", e,
                     traceback.format_exc())
        raise
        # See also good advice, not implemented here, at
        # https://stackoverflow.com/questions/19924104/python-multiprocessing-handling-child-errors-in-parent  # noqa
        # https://stackoverflow.com/questions/6126007/python-getting-a-traceback-from-a-multiprocessing-process/26096355#26096355  # noqa
    # log.critical("process_file: end")


def main() -> None:
    """
    Command-line handler for the ``find_recovered_openxml`` tool.
    Use the ``--help`` option for help.
    """
    parser = ArgumentParser(
        formatter_class=RawDescriptionHelpFormatter,
        description=f"""
Tool to recognize and rescue Microsoft Office OpenXML files, even if they have
garbage appended to them.        

- Rationale: when you have accidentally deleted files from an NTFS disk, and
  they really matter, you should (a) stop what you're doing; (b) clone the disk
  to an image file using "dd" under Linux; (c) perform all subsequent 
  operations on the cloned image (in read-only mode). Those steps might 
  include:
    - ntfsundelete, to find files that the filesystem is still aware of;
    - scalpel, to find files based on their contents.

- Scalpel is great at finding stuff efficiently, but it works best when files
  can be defined by both a start (header) signature and an end (footer)
  signature. However, the Microsoft Office OpenXML file format has a 
  recognizable header, but no standard footer. In these circumstances, Scalpel
  reads up to a certain limit that you specify in its configuration file. (To
  retrieve large Powerpoint files, this limit needs to be substantial, e.g.
  50 Mb or more, depending on your ways of working with Powerpoint.)

- That means that files emerging from a Scalpel search for DOCX/PPTX/XLSX files
  may be
    - false positives, having nothing to do with Office;
    - clean Office files (the least likely category!);
    - Office files with garbage stuck on the end.
    
- The OpenXML file format is just a zip file. If you stick too much garbage on
  the end of a zip file, zip readers will see it as corrupt.  
        
- THIS TOOL detects (and optionally moves) potentially corrupted zipfiles based 
  on file contents, by unzipping the file and checking for "inner" files with
  names like:

        File type       Contents filename signature (regular expression)
        ----------------------------------------------------------------
        DOCX            {DOCX_CONTENTS_REGEX_STR}  
        PPTX            {PPTX_CONTENTS_REGEX_STR}
        XLSX            {XLSX_CONTENTS_REGEX_STR}

- WARNING: it's possible for an OpenXML file to contain more than one of these.
  If so, they may be mis-classified.

- If a file is not immediately readable as a zip, it uses Linux's "zip -FF" to 
  repair zip files with corrupted ends, and tries again.
  
- Having found valid-looking files, you can elect to move them elsewhere.

- As an additional and VERY DANGEROUS operation, you can elect to delete files
  that this tool doesn't recognize. (Why? Because a 450Gb disk might produce
  well in excess of 1.7Tb of candidate files; many will be false positives and
  even the true positives will all be expanded to your file size limit, e.g.
  50 Mb. You may have a problem with available disk space, so running this tool
  regularly allows you to clear up the junk. Use the --run_every option to help 
  with this.)

        """
    )
    parser.add_argument(
        "filename", nargs="+",
        help="File(s) to check. You can also specify directores if you use "
             "--recursive"
    )
    parser.add_argument(
        "--recursive", action="store_true",
        help="Allow search to descend recursively into any directories "
             "encountered."
    )
    parser.add_argument(
        "--skip_files", nargs="*", default=[],
        help="File pattern(s) to skip. You can specify wildcards like '*.txt' "
             "(but you will have to enclose that pattern in quotes under "
             "UNIX-like operating systems). The basename of each file will be "
             "tested against these filenames/patterns. Consider including "
             "Scalpel's 'audit.txt'."
    )
    parser.add_argument(
        "--filetypes", nargs="+", default=FILETYPES,
        help=f"File types to check. Options: {FILETYPES}"
    )
    parser.add_argument(
        "--move_to",
        help="If the file is recognized as one of the specified file types, "
             "move it to the directory specified here."
    )
    parser.add_argument(
        "--delete_if_not_specified_file_type", action="store_true",
        help="If a file is NOT recognized as one of the specified file types, "
             "delete it. VERY DANGEROUS."
    )
    parser.add_argument(
        "--run_repeatedly", type=int,
        help="Run the tool repeatedly with a pause of <run_repeatedly> "
             "seconds between runs. (For this to work well with the move/"
             "delete options, you should specify one or more DIRECTORIES in "
             "the 'filename' arguments, not files, and you will need the "
             "--recursive option.)"
    )
    parser.add_argument(
        "--nprocesses", type=int, default=multiprocessing.cpu_count(),
        help="Specify the number of processes to run in parallel."
    )
    parser.add_argument(
        "--verbose", action="store_true",
        help="Verbose output"
    )
    parser.add_argument(
        "--show_zip_output", action="store_true",
        help="Verbose output from the external 'zip' tool"
    )
    args = parser.parse_args()
    main_only_quicksetup_rootlogger(
        level=logging.DEBUG if args.verbose else logging.INFO,
        with_process_id=True
    )

    # Further argument checks
    if args.move_to and not os.path.isdir(args.move_to):
        raise ValueError(
            f"Destination directory {args.move_to!r} is not a directory")
    if not args.filetypes:
        raise ValueError("No file type to scan for")
    filetypes = [ft.lower() for ft in args.filetypes]
    if any(ft not in FILETYPES for ft in filetypes):
        raise ValueError(f"Invalid filetypes; choose from {FILETYPES}")
    assert shutil.which("zip"), "Need 'zip' tool!"

    # Repeated scanning loop
    while True:
        log.info("Starting scan.")
        log.info("- Looking for filetypes {}", filetypes)
        log.info("- Scanning files/directories {!r}{}",
                 args.filename,
                 " recursively" if args.recursive else "")
        log.info("- Skipping files matching {!r}", args.skip_files)
        log.info("- Using {} simultaneous processes", args.nprocesses)
        if args.move_to:
            log.info("- Moving target files to " + args.move_to)
        if args.delete_if_not_specified_file_type:
            log.info("- Deleting non-target files.")

        # Iterate through files
        pool = multiprocessing.Pool(processes=args.nprocesses)
        for filename in gen_filenames(starting_filenames=args.filename,
                                      recursive=args.recursive):
            src_basename = os.path.basename(filename)
            if any(fnmatch.fnmatch(src_basename, pattern)
                   for pattern in args.skip_files):
                log.info("Skipping file as ordered: " + filename)
                continue
            exists, locked = exists_locked(filename)
            if locked or not exists:
                log.info("Skipping currently inaccessible file: " + filename)
                continue
            kwargs = {
                'filename': filename,
                'filetypes': filetypes,
                'move_to': args.move_to,
                'delete_if_not_specified_file_type':
                    args.delete_if_not_specified_file_type,
                'show_zip_output': args.show_zip_output,
            }
            # log.critical("start")
            pool.apply_async(process_file, [], kwargs)
            # result = pool.apply_async(process_file, [], kwargs)
            # result.get()  # will re-raise any child exceptions
            # ... but it waits for the process to complete! That's no help.
            # log.critical("next")
            # ... https://stackoverflow.com/questions/22094852/how-to-catch-exceptions-in-workers-in-multiprocessing  # noqa
        pool.close()
        pool.join()

        log.info("Finished scan.")
        if args.run_repeatedly is None:
            break
        log.info("Sleeping for {} s...", args.run_repeatedly)
        sleep(args.run_repeatedly)


if __name__ == '__main__':
    main()



#!/usr/bin/env python
# cardinal_pythonlib/django/fields/helpers.py

"""
===============================================================================

    Original code copyright (C) 2009-2021 Rudolf Cardinal (rudolf@pobox.com).

    This file is part of cardinal_pythonlib.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        https://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

===============================================================================

**Helper functions for Django fields.**

"""

from typing import Iterable, Tuple


# =============================================================================
# Field choice assistance
# =============================================================================

def valid_choice(strvalue: str, choices: Iterable[Tuple[str, str]]) -> bool:
    """
    Checks that value is one of the valid option in choices, where choices
    is a list/tuple of 2-tuples (option, description).

    Note that parameters sent by URLconf are always strings
    (https://docs.djangoproject.com/en/1.8/topics/http/urls/)
    but Python is happy with a string-to-integer-PK lookup, e.g.

    .. code-block:: python

        Study.objects.get(pk=1)
        Study.objects.get(pk="1")  # also works

    Choices can be non-string, though, so we compare against a string version
    of the choice.
    """
    return strvalue in [str(x[0]) for x in choices]


def choice_explanation(value: str, choices: Iterable[Tuple[str, str]]) -> str:
    """
    Returns the explanation associated with a Django choice tuple-list.
    """
    for k, v in choices:
        if k == value:
            return v
    return ''



#!/usr/bin/env python
# cardinal_pythonlib/bulk_email/constants.py

"""
===============================================================================

    Original code copyright (C) 2009-2021 Rudolf Cardinal (rudolf@pobox.com).

    This file is part of cardinal_pythonlib.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        https://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

===============================================================================

Constants for the simple bulk e-mail tool.

"""

DB_URL_ENVVAR = "CARDINAL_PYTHONLIB_BULK_EMAIL_DB_URL"

CONTENT_TYPE_MAX_LENGTH = 255
# Can be quite long; see cardinal_pythonlib.httpconst.MimeType
# 255 is the formal limit:
# https://stackoverflow.com/questions/643690/maximum-mimetype-length-when-storing-type-in-db  # noqa

DEFAULT_TIME_BETWEEN_EMAILS_S = 0.5

ENCODING_NAME_MAX_LENGTH = 20  # a guess!
# https://en.wikipedia.org/wiki/Character_encoding

FERNET_KEY_BASE64_LENGTH = 44
# The cryptography.Fernet key is 32 bytes, encrypted via base 64.
# Base 64 encoding: produces 4n/3 characters, so 4 * 32 / 3 = 42.67
# ... padded to multiples of 4, giving 44.

HOSTNAME_MAX_LENGTH = 255

PASSWORD_MAX_LENGTH = 255
PASSWORD_OBSCURING_STRING = "*" * 8

RFC_2822_DATETIME_MAX_LENGTH = 40  # approximately!
# https://datatracker.ietf.org/doc/html/rfc2822#section-3.3
# e.g.
# Wed, 31 Sep 2000 11:29:05 +01:00 (CET)
# 0123456789012345678901234567890123456789
#           1         2         3

USERNAME_MAX_LENGTH = 255
